
#include "some_ut4.h"

#include <experimental/filesystem>
#include <cmath>
namespace fs=std::experimental::filesystem;

// compile possibilities include
// g++ -o $L/file_list.exu file_list.cpp -lstdc++fs
// objcopy -g file_list.exu

// g++ -o y:\file_list.exe file_list.cpp -lstdc++fs

const char psep=
#ifdef _WIN32
  '\\';
#else
  '/';
#endif

#ifndef _WIN32
  #include <dirent.h>
#endif

struct mkw_file_type{
  char *fn;     // file name within a directory
  char *dn;     // full directory name
  double sz;    // size in bytes (whole number but may exceed max 32bit int)
  char f_d;     // 'f' for a file, 'l' for a link, 'd' for a directory entry
  char c14[15]; // 14-digit time stamp
  time_t t;     // another format for the time stamp
  int i;        // index which may be used in sorting decisions
  bool show;    // true if to be shown
};

struct mkw_dir_type{
  char *dn;     // full directory name
  int n;        // number of entries
  int nsub;     // number of entries including sub-directories
  double sz;    // size in bytes (whole number but may exceed max 32bit int)
  double szsub; // size in bytes including sub-directories
#ifndef _WIN32
  dev_t dev;    // device of directory
#endif
};

class order_det{
  public:
  bool doit;    // true of false depending on whether to sort etc in this way
  char fd;      // 'f' for decision based on files, 'd' for directories
  string s;     // description of the sorting criteria
  int(*ord)(const void*, const void*);   // order function
  bool vi;      // show with edit where edit is specified in that variable
  string fil_nam;  // file name to save to
  order_det();
  void set(char fd, const char s[], 
           string sb, const char ext[], string& sfn,
           int(*ord)(const void*, const void*) );
};

order_det::order_det()
{
  doit=false;
  fd=' ';
  s="";
  vi=true;
  fil_nam="";
}

void order_det::set(char f, const char z[], 
                    string sb, const char ext[], string& sfn,
                    int(*o)(const void*, const void*) )
{
  fd=f;
  s=z;
  sfn=sb;
  sfn=sfn+ext+".txt";
  fil_nam=sfn;
  ord=o; 
}

// .. global variables of the maximum space needed for the entries
// .. in any directory, characters are in c[], d[] points to parts
// .. of c[] for the start of each name
char **d=NULL;
char *c=NULL;
int nd_cap=0, nc_cap=0;

// .. global variables of the maximum space needed for the entries
// .. in the command line arguments argv[] and the recursive listings
// .. these are for just the directories encountered
char **ad=NULL, **adu=NULL;
char *adc=NULL;
int nad_cap=0, nadc_cap=0;

// .. global variables of the maximum space needed for the entries
// .. of all the entries encountered
char **df=NULL;
char *cf=NULL;
int ndf_cap=0, ncf_cap=0;

// .. global variable for amount of intermediate output
int verb=1;

// .. global variable for directories to skip
int nskip=0, npskip;
char **dskip=NULL, **dpskip=NULL;

// .. global variable for editor
string edit="gvim";


void usage(const char a[])
{
  int l;
  if(verb>=0)
  {
    cout << "..stopping because: " << a << endl;
    cout << "\nUsage:\n";
    l=strlen(gargv[0]);
    cout << gargv[0] << " directories_to_list (separated by spaces, enter ... for the cwd)\n"
         << setw(l) << "" 
         << " [-skip=dn to skip directory dn (repeat for each one)]\n"
         << setw(l) << "" 
         << " [-pskip=dn to skip directory with a match to dn (may repeat)]\n"
         << setw(l) << "" << " [-v=xx for verbose level]\n"
         << setw(l) << "" << " [-e=editor_to_use] (default editor is gvim)\n" 
         << setw(l) << "" << " [-not_interactive for just save mode]\n" 
         << setw(l) << "" 
         << " [-savea=fna  -saveb=fnb etc to save to named files]\n";
    cout << endl;
  }
  exit(0);
}

char get_character()
{
  char c;

  cin.clear();
  cin >> c;
  cin.ignore(65536, '\n');
  
  if(cin.good()==0)
  {
    return ' ';
  }
  return c;
}

int strcmp_nocase(const char a[], const char b[])
{
  // .. case insensitive comparison
  int i, la, lb, l;
  la=strlen(a);
  lb=strlen(b);
  l=la;
  if(lb<l)l=lb;

  for(i=0; i<l; i++)
  {
    if(tolower(a[i]) < tolower(b[i]) )return -1;
    if(tolower(a[i]) > tolower(b[i]) )return +1;
  }

  if(la<lb)return -1;
  if(la>lb)return +1;
  return 0;
}

int order_alpha(const void *aa, const void *bb)
{
  const char*& a=*(const char**)aa;
  const char*& b=*(const char**)bb;
  return strcmp(a, b);
}

int order_alpha_sep(const void *aa, const void *bb)
{
  const char*& a=*(const char**)aa;
  const char*& b=*(const char**)bb;

  int i, la, lb;
  la=strlen(a);
  lb=strlen(b);
  for(i=0; i<la && i<lb; i++)
  {
    if(a[i]==psep && b[i]!=psep)return -1;
    if(a[i]!=psep && b[i]==psep)return +1;
    if(a[i]<b[i])return -1;
    if(a[i]>b[i])return +1;
  }
  if(la<lb)return -1;
  if(la>lb)return +1;
  return 0;
}

int order_pf_timestamp(const void *aa, const void *bb)
{
  const mkw_file_type& a=**(const mkw_file_type**)aa;
  const mkw_file_type& b=**(const mkw_file_type**)bb;

  // .. order by the timsstamp

  int k;
  k=strcmp(a.c14, b.c14);
  if(k<0)return +1;
  if(k>0)return -1;

  k=order_alpha_sep(&a.dn, &b.dn);
  if(k!=0)return k;
  k=order_alpha_sep(&a.fn, &b.fn);
  return k;
}

int order_pf_filesize(const void *aa, const void *bb)
{
  const mkw_file_type& a=**(const mkw_file_type**)aa;
  const mkw_file_type& b=**(const mkw_file_type**)bb;

  // .. order by the file size first

  int k;
  if(a.sz<b.sz)return +1;
  if(a.sz>b.sz)return -1;

  k=strcmp(a.c14, b.c14);
  if(k<0)return +1;
  if(k>0)return -1;

  k=order_alpha_sep(&a.dn, &b.dn);
  if(k!=0)return k;
  k=order_alpha_sep(&a.fn, &b.fn);
  return k;
}

int order_pf_alpha(const void *aa, const void *bb)
{
  const mkw_file_type& a=**(const mkw_file_type**)aa;
  const mkw_file_type& b=**(const mkw_file_type**)bb;

  // .. alphabetical by the full file name

  int k;
  k=order_alpha_sep(&a.dn, &b.dn);
  if(k!=0)return k;
  k=order_alpha_sep(&a.fn, &b.fn);
  return k;
}

int order_pf_loc_alpha(const void *aa, const void *bb)
{
  const mkw_file_type& a=**(const mkw_file_type**)aa;
  const mkw_file_type& b=**(const mkw_file_type**)bb;

  // .. alphabetical by the local file name first
  // .. followed by size, followed by timestamp followed by directory name

  int k;
  k=order_alpha_sep(&a.fn, &b.fn);
  if(k!=0)return k;

  if(a.sz<b.sz)return 1;
  if(a.sz>b.sz)return -1;

  k=strcmp(a.c14, b.c14);
  if(k!=0)return k;

  return strcmp(a.dn, b.dn);
}

int order_pd_numb(const void *aa, const void *bb)
{
  const mkw_dir_type& a=**(const mkw_dir_type**)aa;
  const mkw_dir_type& b=**(const mkw_dir_type**)bb;

  if(a.n<b.n)return +1;
  if(a.n>b.n)return -1;
  return strcmp(a.dn, b.dn);
}

int order_pd_subdir_numb(const void *aa, const void *bb)
{
  const mkw_dir_type& a=**(const mkw_dir_type**)aa;
  const mkw_dir_type& b=**(const mkw_dir_type**)bb;

  if(a.nsub<b.nsub)return +1;
  if(a.nsub>b.nsub)return -1;
  return strcmp(a.dn, b.dn);
}

int order_pd_size(const void *aa, const void *bb)
{
  const mkw_dir_type& a=**(const mkw_dir_type**)aa;
  const mkw_dir_type& b=**(const mkw_dir_type**)bb;

  if(a.sz<b.sz)return +1;
  if(a.sz>b.sz)return -1;
  return strcmp(a.dn, b.dn);
}

int order_pd_subdir_size(const void *aa, const void *bb)
{
  const mkw_dir_type& a=**(const mkw_dir_type**)aa;
  const mkw_dir_type& b=**(const mkw_dir_type**)bb;

  if(a.szsub<b.szsub)return +1;
  if(a.szsub>b.szsub)return -1;
  return strcmp(a.dn, b.dn);
}

int order_alpha_nocase(const void *aa, const void *bb)
{
  const char*& a=*(const char**)aa;
  const char*& b=*(const char**)bb;
  return strcmp_nocase(a, b);
}

int order_pdir(const void *aa, const void *bb)
{
  const mkw_dir_type& a=**(const mkw_dir_type**)aa;
  const mkw_dir_type& b=**(const mkw_dir_type**)bb;
  return order_alpha_sep(&a.dn, &b.dn);
}

void merge_arrays(void* va1,  int n1,  
                  void* va2,  int n2,  
                  void* vb, 
                  int m,  int (*f)(const void*,  const void*))
{
  int i1=0,  i2=0,  j;
  // .. recast the pointers
  char *a1=reinterpret_cast<char*>(va1);
  char *a2=reinterpret_cast<char*>(va2);
  char *b=reinterpret_cast<char*>(vb);

  for(j=0; j<n1+n2; j++)
  {
    if(i1==n1)  // just use a2[] entries
    {
      memcpy(&b[j*m],  &a2[i2*m],  m);
      i2++; 
      continue;
    }

    if(i2==n2)  // just use a1[] entries
    {
      memcpy(&b[j*m],  &a1[i1*m],  m);
      i1++; 
      continue;
    }

    if(f(&a1[i1*m],  &a2[i2*m])<0) // select smallest
    {
      memcpy(&b[j*m],  &a1[i1*m],  m);
      i1++;
    }
    else
    {
      memcpy(&b[j*m],  &a2[i2*m],  m);
      i2++;
    }
  }
}

void msort(void* y,  int n,  int m,  
           int (*f)(const void*,  const void*))
{

  int i,  l;
  if(n<=0)return;
  char *x=reinterpret_cast<char*>(y);

  // .. allocate temporary space 
  char *tem=new char[n*m];
  memcpy(tem, x, n*m);

  l=1;
  while(l<n)
  {
    // merge the blocks of length <= l
    for(i=0; i<n; i=i+2*l)
    {
      if(i+l>=n)break;  
      int l2=l;
      if(i+2*l>n)l2=n-(i+l);
      merge_arrays(&x[i*m],  l,  
                   &x[(i+l)*m],  l2,  &tem[i*m], 
                   m, f);
    }

    // .. copy tem to x
    memcpy(x, tem, n*m);

    // double l for the next step
    l=2*l;
  }

  // .. remove the temporary space
  delete[] tem; 
}

void increase_cap_file_entries(int& n, mkw_file_type*& a)
{
  int i;

  // .. take a copy of a[] as b[]
  mkw_file_type *b=new mkw_file_type[n];

  for(i=0; i<n; i++)
  {
    mkw_file_type& y=a[i];
    mkw_file_type& z=b[i];
    z.fn=y.fn;  
    z.dn=y.dn;  
    z.sz=y.sz;  
    z.f_d=y.f_d;
    mstrncpy(z.c14, y.c14, 14);
    z.c14[14]=0;  
    z.t=y.t;
    z.i=y.i;
  }

  // .. resize a[] and then copy entries in b[]
  a=new mkw_file_type[2*n];
  for(i=0; i<n; i++)
  {
    mkw_file_type& z=a[i];
    mkw_file_type& y=b[i];
    z.fn=y.fn;  
    z.dn=y.dn;  
    z.sz=y.sz;  
    z.f_d=y.f_d;
    mstrncpy(z.c14, y.c14, 14);
    z.c14[14]=0;  
    z.t=y.t;
    z.i=y.i;
  }

  // .. remove the temporary array b[] and reset n
  delete[] b;
  n=2*n;
}

void increase_cap_dir_entries(int& n, mkw_dir_type*& a)
{
  int i;

  // .. take a copy of a[] as b[]
  mkw_dir_type *b=new mkw_dir_type[n];

  for(i=0; i<n; i++)
  {
    mkw_dir_type& y=a[i];
    mkw_dir_type& z=b[i];
    z.dn=y.dn;  
    z.nsub=y.nsub;
    z.sz=y.sz;
    z.szsub=y.szsub;
#ifndef _WIN32
    z.dev=y.dev;
#endif
  }

  // .. resize a[] and then copy entries in b[]
  a=new mkw_dir_type[2*n];
  for(i=0; i<n; i++)
  {
    mkw_dir_type& z=a[i];
    mkw_dir_type& y=b[i];
    z.dn=y.dn;  
    z.nsub=y.nsub;
    z.sz=y.sz;
    z.szsub=y.szsub;
#ifndef _WIN32
    z.dev=y.dev;
#endif
  }

  // .. remove the temporary array b[] and reset n
  delete[] b;
  n=2*n;
}

void increase_cap(int nc_above, int& nc_cap, char*& c, 
                  int nd_above, int& nd_cap, char**& d, const char des[])
{
  int i, j;

  if(nc_above>=nc_cap && verb>=1)
  {
    cout << des << ": nc_above=" << nc_above 
         << ", nd_above=" << nd_above;
    cout << " ..need to increase nc_cap=" << nc_cap << endl;
  }
  if(nd_above>=nd_cap && verb>=1)
  {
    cout << des << ": nc_above=" << nc_above 
         << ", nd_above=" << nd_above;
    cout << " ..need to increase nd_cap=" << nd_cap << endl;
  }
  
  // .. get the lengths of the d[] entries so far
  int *zl=new int[nd_above];
  if(zl==NULL)
  {
    cout << "failed to allocate space for zl[]" << endl;
    usage("cannot continue");
  }

  for(i=0; i<nd_above-1; i++)zl[i]=strlen(d[i]);


  char *c2;

  // .. increase nc_cap case
  if(nc_above>=nc_cap)
  {
    // .. take of copy of the current c[] as c2
    c2=new char[nc_cap*sizeof(char)];
    if(c2==NULL)
    {
      cout << "failed to allocate space for c2[]" << endl;
      usage("cannot continue");
    }

    memcpy(c2, c, nc_cap*sizeof(char));

    // .. replace c by double the size version
    delete[] c;
    c=new char[2*nc_cap];
    if(c==NULL)
    {
      cout << "failed to allocate space for c[]" << endl;
      usage("cannot continue");
    }

    memcpy(c, c2, nc_cap);
    nc_cap=2*nc_cap;

    // .. re-construct d[] using the new c[]
    j=0;
    for(i=0; i<nd_above-1; i++)
    {
      d[i]=&c[j];
      j=j+zl[i]+1;
    }
    d[nd_above-1]=&c[j];

    // .. remove temporary array c2
    delete[] c2;
  }
  
  // .. increase nd_cap case
  if(nd_above>=nd_cap)
  {
    delete[] d;
    d=new char*[2*nd_cap];
    if(d==NULL)
    {
      cout << "failed to allocate space for d[]" << endl;
      usage("cannot continue");
    }

    nd_cap=2*nd_cap;

    // .. re-construct new d[]
    j=0;
    for(i=0; i<nd_above-1; i++)
    {
      d[i]=&c[j];
      j=j+zl[i]+1;
    }
    d[nd_above-1]=&c[j];
  }

  // .. remove temporary array zl
  delete[] zl;
}

void get_in_this_directory(const char p[], int& n, int& nc)
{
  string s;
  int l, nc1;

  n=0;
  nc=0;
  d[0]=c;
  //for(const auto& e: fs::recursive_directory_iterator(p) )

  try{
    for(const auto& e: fs::directory_iterator(p, 
          fs::directory_options::skip_permission_denied | fs::directory_options::none) )
    {
      const fs::path& pa=e.path();
      s=pa.string();
      s=pa.string();
      l=s.length();
      l=s.length();
  
      nc1=nc+l+1;
      if(nc1>=nc_cap || n>=nd_cap) 
        increase_cap(nc1, nc_cap, c, n+1, nd_cap, d, "within_dir_storage");
      mstrncpy(d[n], &s[0], l);
      d[n][l]=0;
      n++;
      d[n]=&c[nc1];
      nc=nc1;
    }
  }
  catch(fs::filesystem_error &e)
  {
    if(verb>=1)
      cout << "directory skipped because of exception with directory=" 
           << p << " " << e.what() << endl;
  }
}

bool same_filesystem(const char a[], const char b[])
{
  if(psep=='\\')return true;

  // .. on a linux system check if a[] and b[] are on the same filesystem

#ifndef _WIN32
  struct stat finfo_a, finfo_b;
  int ia=lstat(a, &finfo_a);
  int ib=lstat(b, &finfo_b);
  if(finfo_a.st_dev != finfo_b.st_dev)
  {
    if(verb>=1)
      cout << "different file systems: " << a << " and " << b << endl;
    return false;
  }
#endif

  return true;
}

void file_details3(const char fname[], char c14age[], time_t& age,
                   int& bytes, int& isdir)
{
  struct stat finfo;
  unsigned short mty;
  isdir=0;
  int ij=stat(fname, &finfo);

  // .. deal with cases when the file is not accessed
  if (ij!=0)
  {
    mstrcpy(c14age, "00000000000000");
    bytes=0;
    return;
  }
  age=finfo.st_mtime;
  mty=finfo.st_mode;

  // .. alter any dates out of range to prevent failure
  if(age<0)age=0;

  // .. indictate if a directory
  if(S_ISDIR(mty))isdir=1;

  struct tm g;
  time_t_to_struct_tm(age, g);
  stringstream sstem;
  string stem;

  sstem.clear();
  sstem.str("");
  sstem << 1900+g.tm_year << setw(2) << 1+g.tm_mon << setw(2) << g.tm_mday
  << setw(2) << g.tm_hour << setw(2) << g.tm_min << setw(2) << g.tm_sec;
  stem=sstem.str();
  mstrcpy(c14age, stem.c_str());

  int i;
  for (i=0; i<14; i++)if (!isdigit(c14age[i]))c14age[i]='0';

  bytes=int(finfo.st_size);
}

void get_more_details(int m, char *a[], mkw_file_type*& f, 
                      int& nd, mkw_dir_type*& d)
{
  f=new mkw_file_type[m+1];

  int i, j, l, bytes, isdir, l1;
  char *c;
  string s, sl;
  fs::path p, pl;

  // .. loop through the entries to classify, get the timestamp and size
  for(i=0; i<m; i++)
  {
    mkw_file_type& y=f[i];
    y.show=true;

    c=a[i];
    s=c;
    p=c;
    l=strlen(c);

    // .. extract the directory name and file name from c
    // .. when the entry is not a directory
    y.dn=c;
    y.fn=blank;
    for(j=l-1; j>=0; j--)
    {
      if(c[j]==psep)
      {
        c[j]=0;
        y.fn=&c[j+1];
        break;
      }
    }

    // .. get time stamp
    file_details3(&s[0], y.c14, y.t, bytes, isdir);
    if(fs::is_symlink(&s[0])==true)
    {
      y.f_d='l';
      pl=read_symlink(p);
      sl=pl.string();
      y.sz=s.length();
      continue;
    }

    if(fs::is_directory(&s[0])==true)
    {
      y.f_d='d';
      y.sz=0;
    }
    else
    {
      // .. get the file size which could be large
      y.f_d='f';
      y.sz=0;
      if(fs::is_regular_file(&s[0])==true)
        y.sz=fs::file_size(p);
    }
  }

  // .. deduce the number of directory groups as nd
  nd=0;
  if(m>0)nd++;

  for(i=1; i<m; i++)
  {
    mkw_file_type& y1=f[i-1];
    mkw_file_type& y=f[i];
    if(strcmp(y.dn, y1.dn)!=0) nd++;
  }
  d=new mkw_dir_type[nd+1];

  // .. set the name, number and size with a directory only
  j=0;
  if(m>0)
  {
    mkw_file_type& y=f[0];
    mkw_dir_type& z=d[j];
    z.dn=y.dn;
    z.n=0;
    z.sz=0;
    z.nsub=0;
    z.szsub=0;
    j++; 
  }
  for(i=1; i<m; i++)
  {
    mkw_file_type& y1=f[i-1];
    mkw_file_type& y=f[i];
    mkw_dir_type& z1=d[j-1];
    if(strcmp(y.dn, y1.dn)!=0)
    {
      mkw_dir_type& z=d[j];
      z.dn=y.dn;
      z.n=0;
      z.sz=0;
      z.nsub=0;
      z.szsub=0;
      j++;
    }
    else
    {
      z1.n++;
      z1.sz += y.sz;
    }
  }

  // .. create an array of pointers and sort these as a step
  // .. towards seting the nsub and szsub entries
  mkw_dir_type **pd=new mkw_dir_type*[nd+1];
  for(i=0; i<nd; i++)pd[i]=&d[i];
  msort(pd, nd, sizeof(pd[0]), order_pdir);

  for(i=0; i<nd; i++)
  {
    mkw_dir_type& z=*pd[i];
    z.nsub=z.n;
    z.szsub=z.sz;
    l=strlen(z.dn);

    // .. search forward until no longer a sub-directory
    for(j=i+1; j<nd; j++)
    {
      mkw_dir_type& z1=*pd[j];
      l1=strlen(z1.dn);
      if(l1<l)break;
      if(strncmp(z1.dn, z.dn, l)!=0)break;
      if(z1.dn[l]!=psep)break;
      z.nsub += (z1.n+1);
      z.szsub += z1.sz; 
    }
  }
}

void show_in_editor(const char fn[])
{
  string s;
  s=edit;
  s=s+"  "+fn;
  system(&s[0]);
}

template<class T>
void show_save_f_gen(T& O, order_det& o, mkw_file_type *pf[], int n)
{
  int i;
  O << setprecision(0) << fixed;
  for(i=0; i<n; i++)
  {
    mkw_file_type& y=*pf[i];
    if(y.show==false)continue;

    O << y.c14 << " " << setw(12) << y.sz;
    O << " " << setw(3);
    if(y.f_d=='f')O << "";
    if(y.f_d=='d')O << "<d>";
    if(y.f_d=='l')O << "<l>";
    O << " " << y.dn << psep << y.fn << endl;
  }
}

template<class T>
void show_save_d_gen(T& O, order_det& o, mkw_dir_type *pd[], int n)
{
  int i;
  O << setprecision(0) << fixed;
  for(i=0; i<n; i++)
  {
    mkw_dir_type& y=*pd[i];
    O << " " << setw(12) << y.szsub << " bytes"
      << " " << setw(12) << y.sz << " bytes"
      << " " << setw(9) << y.nsub
      << " " << setw(9) << y.n;

    O << " " << y.dn << endl;
  }
}

void show_save_f(order_det& o, mkw_file_type *pf[], int n)
{
  ofstream O;
  string s;
  char ck;

  s=o.fil_nam;
  if(s.length()>0)
  {
    if(verb>=0)cout << "...creating " << s << endl;
    O.open(s.c_str());
    show_save_f_gen(O, o, pf, n);
    O.close();

    show_in_editor(&s[0]);
    cout << "press any letter a-z,A-Y and the file will be removed\n";
    cout << "pressing Z removes the file and exits the program\n";
    ck=get_character();
    if( ck==' ' || ('a'<=ck && ck<='z') || ('a'<=ck && ck<='Z') )
    {
      cout << "removing " << s << endl;
      remove(&s[0]);
    }
    if(ck=='Z')exit(0);

    return;
  }

  show_save_f_gen(cout, o, pf, n);
}

void just_save_f(order_det& o, mkw_file_type *pf[], int n)
{
  ofstream O;
  string s;

  s=o.fil_nam;
  if(s.length()>0)
  {
    if(verb>=0)cout << "...creating " << s << endl;
    O.open(s.c_str());
    show_save_f_gen(O, o, pf, n);
    O.close();
  }
}

void show_save_d(order_det& o, mkw_dir_type *pd[], int n)
{
  ofstream O;
  string s;
  char ck;

  s=o.fil_nam;
  if(s.length()>0)
  {
    if(verb>=0)cout << "...creating " << s << endl;
    O.open(s.c_str());
    show_save_d_gen(O, o, pd, n);
    O.close();

    show_in_editor(&s[0]);
    cout << "press any letter a-z,A-Z and the file will be removed\n";
    ck=get_character();
    if( ('a'<=ck && ck<='z') || ('a'<=ck && ck<='Z') )remove(&s[0]);

    return;
  }

  show_save_d_gen(cout, o, pd, n);
}

void just_save_d(order_det& o, mkw_dir_type *pd[], int n)
{
  ofstream O;
  string s;

  s=o.fil_nam;
  if(s.length()>0)
  {
    if(verb>=0)cout << "...creating " << s << endl;
    O.open(s.c_str());
    show_save_d_gen(O, o, pd, n);
    O.close();
  }
}

string temp_dir()
{
  // .. set the temporary directory to be used
  string s; 
  if(exists_dir("/home/latex"))
  {
    s="/home/latex/";
    return s;
  }

  char *p;
  p=getenv("TEMP");
  if(p!=NULL)
  {
    s=p;
    s=s+psep;
    return s;
  }

  p=getenv("TMP");
  if(p!=NULL)
  {
    s=p;
    s=s+psep;
    return s;
  }

  s="";
  return s;
}

bool skip_dir(const char z[])
{
  // .. return true if z[] exactly matches something in dskip[]
  int i, k;
  for(i=0; i<nskip; i++)
    if(strcmp(z, dskip[i])==0)
    {
      if(verb>=1)cout << "...as requested skipping " << z << endl;
      return true;
    }
  return false;
}

bool pskip_dir(const char z[])
{
  // .. return true if z[] has a match with a name in dpskip[]
  int i, k;
  for(i=0; i<npskip; i++)
  {
    if(psep=='/')k=mkwfind(z, dpskip[i]);
    else         k=mkwfind_nocase(z, dpskip[i]);
    if(k>=0)
    {
      if(verb>=1)cout << "...as requested skipping " << z << endl;
      return true;
    }
  }
  return false;
}

void sift(int n, mkw_file_type *px[])
{
  cout << "Specify all the strings, separated by spaces, to be shown next,\n"
       << "e.g. .m .docx .xlsx .tex\n"
       << "to restrict to names which contain at least 1 of these strings,\n"
       << "/////  to reset to show everything\n";
  cout << "Enter your choice: ";

  string s;
  int i, j, k, l, m;

  cin >> s;
  cin.clear();
  cin.ignore(65536, '\n');

  // .. get the array of strings as a[];
  l=s.length();
  char **a=new char*[l+1];

  k=0;
  if(s[0]!=' ')
  {
    a[k]=&s[0];
    k++;
  }
  for(i=1; i<l; i++)
  {
    if(s[i-1]==' ' && s[i]!=' ')
    {
      a[k]=&s[i];
      k++;
    }
  }

  if(strcmp(&s[0], "/////")==0)k=0;

  for(i=0; i<l; i++)
    if(s[i]==' ')s[i]=0;
  if(k==0)cout << k << " strings entered, all files will be shown." << endl;
  if(k==1)cout << k << " string entered" << endl;
  if(k>1)cout << k << " strings entered" << endl;

  // .. set the .show entry of entries in px[]
  m=n;

  // .. nothing given, set to show all (again)
  if(l>0 && k==0)
  {
    for(i=0; i<n; i++)
    {
      mkw_file_type& y=*px[i];
      y.show=true;
    }
  }

  if(k>0)m=0;
  for(i=0; i<n && k>0; i++)
  {
    mkw_file_type& y=*px[i];
    y.show=false;
    for(j=0; j<k; j++)
    {
      if(mkwfind(y.fn, a[j])>=0)
      {
        y.show=true;
        m++;
        break;
      }
    }
  }
  cout << m << " file names to be displayed." << endl;
  
  delete[] a;
}

int main(int argc, char *argv[])
{
  #include "some_ut_in_main4.h"

  // .. set starting size to store the names in a single directory 
  nc_cap=1024*1024;
  nd_cap=65536;

  // .. set starting size to store all the names of the directories 
  ncf_cap=2*nc_cap;
  ndf_cap=2*nd_cap;

  int i, k, l, n, nc, kc, kd, l_cwd;
  char *da, *p;

  // .. get the number of chars in command line arguments as kc
  // .. taking care of the cwd shortcut
  const int nc_cwd=4196;
  char c_cwd[nc_cwd];
  get_cwd(c_cwd, nc_cwd);
  l_cwd=strlen(c_cwd);

  kc=0;
  for(i=1; i<argc; i++)
  {
    p=argv[i];
    l=strlen(p);
    if(strcmp(p, "...")==0)
    {
      l=l_cwd;
    }
    kc=kc+l+1; 
  }

  // .. determine if a verbose level other than the default is
  // .. to be used and set it
  for(i=1; i<argc; i++)
    if(strncmp(argv[i], "-v=", 3)==0)
    {
      if(strlen(argv[i])>3)
        verb=atoi(&argv[i][3]);
      break;
    }
    

  // .. set starting size to store the names with recursive version
  // .. and this must at least be enough for the command arguments
  nadc_cap=0;
  nad_cap=0;
  for(i=2; i<=2048; i=2*i)
  {
    k=i*nc_cap;
    kd=i*nd_cap;
    if(k>=kc && kd>=argc)
    {
      nadc_cap=k;
      nad_cap=kd;
      break;
    }
  }
  if(nadc_cap==0)
  {
    if(verb>=0)
      cout << "from analysing the command line arguments, k=" 
           << k << ", kd=" << k << endl;
    usage("these are too large to use for nadc_cap and/or nad_cap");
  }

  // .. allocate the starting space
  c=new char[nc_cap];
  d=new char*[nd_cap];

  adc=new char[nadc_cap];
  ad=new char*[nad_cap];
  adu=new char*[nad_cap];

  cf=new char[ncf_cap];
  df=new char*[ndf_cap];

  if(verb>=2)
  {
    cout << "at the start nc_cap=" << nc_cap << ", nd_cap=" << nd_cap << endl;
    cout << "at the start nadc_cap=" << nadc_cap << ", nad_cap=" << nad_cap << endl;
    cout << "at the start ncf_cap=" << ncf_cap << ", ndf_cap=" << ndf_cap << endl;
  }

  int m, mc, mu, mcu;

  // .. get the directories on the command line uniquely
  if(argc<2)usage("argc<2");
  m=0;
  mc=0;
  for(i=1; i<argc; i++)
  {
    p=argv[i];
    if(strcmp(p, "...")==0)p=c_cwd;

    if(exists_dir(p)==0)continue;
    l=strlen(p);
    ad[m]=&adc[mc];
    mstrncpy(ad[m], p, l);
    ad[m][l]=0;
    mc=mc+l+1;
    m++;
  }
  if(m==0)
    usage("m==0, no directories that exist as a command line arguments");

  if(psep=='/')
    msort(ad, m, sizeof(ad[0]), order_alpha);
  else
    msort(ad, m, sizeof(ad[0]), order_alpha_nocase);

  mu=0;
  adu[mu]=ad[0];
  mu++;
  for(i=1; i<m; i++)
  {
    p=ad[i];
    if(strcmp(p, adu[mu-1])==0)continue;
    if(psep=='\\')
    {
      if(strcmp_nocase(p, adu[mu-1])==0)continue;
    }
    l=strlen(p);
    adu[mu]=p;
    mu++;
  }
  if(verb>=0)
  {
    cout << "m=" << m << " directories on the command line";
    if(m>1)
      cout << ", mu=" << mu << " unique entries." << endl;
    else
      cout << endl;
  }


  // .. put adu[] entries into ad[] and re-do array ac[]
  char *adc_copy=new char[mc];

  int id, nad, nac, nc1, mf, ndf, ncf, ncf1, ndir;

  nad=0;
  nac=0;
  for(i=0; i<mu; i++) 
  {
    p=adu[i];
    l=strlen(p);
    ad[nad]=&adc_copy[nac];
    mstrncpy(ad[nad], p, l);
    ad[nad][l]=0;
    nad++;
    nac=nac+l+1;
  }

  memcpy(adc, adc_copy, nac);
  nac=0;
  for(i=0; i<nad; i++)
  {
    p=ad[i];
    l=strlen(p);
    ad[i]=&adc[nac];
    nac=nac+l+1;
  }
  delete[] adc_copy;

  // .. get any possible directories to skip
  npskip=0;
  nskip=0;
  dskip=new char*[argc+1];
  dpskip=new char*[argc+1];
  for(i=1; i<argc; i++)
  {
    da=argv[i];
    l=strlen(da);
    if(strncmp(da, "-pskip=", 7)==0 && l>=8)
    {
      dpskip[npskip]=&da[7];
      npskip++;
      continue;
    }
    if(strncmp(da, "-skip=", 6)==0 && l>=7)
    {
      dskip[nskip]=&da[6];
      nskip++;
      continue;
    }
  }
  if(nskip>0)
    msort(dskip, nskip, sizeof(dskip[0]), order_alpha);
  if(npskip>0)
    msort(dpskip, npskip, sizeof(dpskip[0]), order_alpha);

  // .. loop to go through the nad directories in ad[]
  // .. id indicates the one being considered
  // .. nad increases as new directories are enountered
  // .. mf indicates the position in cf[] array for the things collected

  id=0;
  mf=0;
  ncf=0;
  do
  {
    p=ad[id];
    l=strlen(p);

    // .. add the entry to the end of the array
    ncf1=ncf+l+2;
    if(ncf1>=ncf_cap || mf+1>=ndf_cap)
      increase_cap(ncf1, ncf_cap, cf, mf+1, ndf_cap, df, "entry_names_storage");

    df[mf]=&cf[ncf];
    mstrncpy(df[mf], p, l);
    df[mf][l]='/';
    df[mf][l+1]=0;

    mf++;
    ncf=ncf1;

    get_in_this_directory(p, n, nc);
    if(verb>=2)
    {
      cout << "id=" << id << ", nad=" << nad 
           << ", p=" << p << ", n=" << n << endl;
    }

    for(i=0; i<n; i++)
    {
      // .. put the entry into cf[] and df[], 
      // .. first check that the arrays are large enough
      l=strlen(d[i]);
      ncf1=ncf+l+1;
      if(ncf1>=ncf_cap || mf+1>=ndf_cap)
        increase_cap(ncf1, ncf_cap, cf, mf+1, ndf_cap, df, "entry_names_storage");


      // .. add the entry to the end of the array
      df[mf]=&cf[ncf];
      mstrncpy(df[mf], d[i], l);
      df[mf][l]=0;

      mf++;
      ncf=ncf1;

      // .. next determine if the entry d[i] is a directory to add to ad[]

      // .. skip links first to cope with links which just
      // .. refer to themselves which leads to an error
      if(fs::is_symlink(d[i])==true)continue;
      if(fs::is_directory(d[i])==false)continue;

      // .. only add directory if p and d[i] are on the same filesystem
      // .. in the case of a linux filesystem which has mount points
      if(same_filesystem(ad[id], d[i])==false)continue;

      // .. check if directory has been requested to be skipped
      if(skip_dir(d[i]))continue;
      if(pskip_dir(d[i]))continue;

      // .. determine the new capacity needed first for the ad[], adc[]
      // .. storage of the directories and increase it if necessary
      nc1=nac+l+1;
      if(nc1>=nadc_cap || nad>=nad_cap) 
        increase_cap(nc1, nadc_cap, adc, nad+1, nad_cap, ad, "dir_names_storage");

      // .. add the directory to the end of the array
      ad[nad]=&adc[nac];
      mstrncpy(ad[nad], d[i], l);
      ad[nad][l]=0;
      nad++;
      nac=nc1;
    }

    id++;
    
  }while(id<nad);

  if(verb>=0)
  {
    cout << "\n...mf=" << mf << " entries found, now getting more details" << endl; 
  }
  mkw_file_type *f_mkw;
  mkw_dir_type *d_mkw;

  get_more_details(mf, df, f_mkw, ndir, d_mkw);
  if(verb>=0)
    cout << "...details obtained, mf=" << mf 
         << ", number of directories is ndir=" << ndir << endl;

  // .. allocate space for pointers which are used in sorting
  mkw_file_type **pf=new mkw_file_type*[mf+1];
  mkw_dir_type **pd=new mkw_dir_type*[ndir+1];

  // .. set base name of any files to be created
  char c20[21];
  string sbase, sfn;
  sbase=temp_dir();
  accurate_time(c20);
  sbase=sbase+"file_listing_"+c20;
  

  // .. set-up the possible processing actions
  order_det by_glob_fil_time, by_glob_fil_size,
    by_glob_fil_alpha, by_loc_fil_alpha, 
    by_dir_numb, by_dir_size, by_subdir_numb, by_subdir_size;

  by_glob_fil_time.set('f', "Order by file timestamp", 
    sbase, "_timestamp", sfn, order_pf_timestamp);

  by_glob_fil_size.set('f', "Order by file size", 
    sbase, "_filesize", sfn, order_pf_filesize);

  by_glob_fil_alpha.set('f', "Alphabetical by full name", 
    sbase, "_full_name", sfn, order_pf_alpha);

  by_loc_fil_alpha.set('f', "Alphabetical by local file name", 
    sbase, "_local_name", sfn, order_pf_loc_alpha);

  by_dir_numb.set('d', "Order of directories by the number of files in them", 
    sbase, "_number_in_dir", sfn, order_pd_numb);
  
  by_dir_size.set('d', "Order of directories by the size of the files in them", 
    sbase, "_size_in_dir", sfn, order_pd_size);

  by_subdir_numb.set('d', 
    "Order of directories by the number of the files also counting sub-dir",
    sbase, "_number_in_all_subdir", sfn, order_pd_subdir_numb);
  
  by_subdir_size.set('d', 
    "Order of directories by the size of files in them and in sub-dir", 
    sbase, "_size_in_all_subdir", sfn, order_pd_subdir_size);

  order_det ord_dec[]={
    by_glob_fil_time, by_glob_fil_size,
    by_glob_fil_alpha, by_loc_fil_alpha, 
    by_dir_numb, by_dir_size, by_subdir_numb, by_subdir_size};
  int nact=sizeof(ord_dec)/sizeof(ord_dec[0]);


  int j;
  char act;
  bool b_interactive=true;

  // .. determine if not interactive
  for(j=1; j<argc; j++)
    if(mkwfind(argv[j], "-not_interactive")>=0){b_interactive=false; break;}

  // .. possibly reset the editor
  for(j=1; j<argc; j++)
    if(strncmp(argv[j], "-e=", 3)==0 && strlen(argv[j])>=4)
    {
      edit=&argv[j][3];
      break;
    }
    

  // .. interactive loop relating to how to sort and what to show
  do
  {
    if(b_interactive==false)break;

    // .. reset all to not process
    for(j=0; j<nact; j++)
    {
      order_det& o=ord_dec[j];
      o.doit=false;
    }

    for(j=0; j<nact; j++)
    {
      order_det& o=ord_dec[j];
      cout << char('a'+j) << "  " << o.s << endl;
    }
    cout << "?  to get prompt to only show names containing a given strings\n";
    cout << "Z  to leave this loop" << endl;
    cout << "enter one of these letters ";
    act=get_character();
    cout << endl;
    if(act=='?')
    {
      for(j=0; j<mf; j++) pf[j]=&f_mkw[j];
      sift(mf, pf);
    }
    if(act=='Z')break;
    if(act<'a')continue;
    if(act>=char('a'+nact))continue;
    j=act-'a';
    order_det& o=ord_dec[j];
    o.doit=true;

    for(i=0; i<nact; i++)
    {
      order_det& o=ord_dec[i];
      if(o.fd=='f' && o.doit)
      {
        for(j=0; j<mf; j++) pf[j]=&f_mkw[j];
        if(verb>=0)cout << "...sorting to do " << o.s << endl;
        msort(pf, mf, sizeof(pf[0]), o.ord);
        show_save_f(o, pf, mf);
      }
      if(o.fd=='d' && o.doit)
      {
        for(j=0; j<ndir; j++) pd[j]=&d_mkw[j];
        if(verb>=0)cout << "...sorting to do " << o.s << endl;
        msort(pd, ndir, sizeof(pd[0]), o.ord);
        show_save_d(o, pd, ndir);
      }
    }
  }while(act!='Z');

  // .. just save files case
  if(b_interactive==false)
  {
    if(verb>=1)
      cout << ".. just saving the listing files depending on\n" 
           << ".. -savea=file_name -saveb=file_name etc command line arguments" 
           << endl;
    // .. reset all to not process
    for(j=0; j<nact; j++)
    {
      order_det& o=ord_dec[j];
      o.doit=false;
    }

    for(i=1; i<argc; i++)
    {
      da=argv[i];
      if(strncmp(da, "-save", 5)!=0)continue;
      l=strlen(da);
      if(l<8)continue;
      if(da[5]<'a' || da[5]>=char('a'+nact))continue;
      if(da[6]!='=')continue;
      j=da[5]-'a';
      order_det& o=ord_dec[j];
      o.fil_nam=temp_dir();
      o.fil_nam=o.fil_nam+"file_listing_"+&da[7];
      o.doit=true;
    }

    for(i=0; i<nact; i++)
    {
      order_det& o=ord_dec[i];
      if(o.fd=='f' && o.doit)
      {
        for(j=0; j<mf; j++) pf[j]=&f_mkw[j];
        if(verb>=0)cout << "...sorting to do " << o.s << endl;
        msort(pf, mf, sizeof(pf[0]), o.ord);
        just_save_f(o, pf, mf);
        continue;
      }
      if(o.fd=='d' && o.doit)
      {
        for(j=0; j<ndir; j++) pd[j]=&d_mkw[j];
        if(verb>=0)cout << "...sorting to do " << o.s << endl;
        msort(pd, ndir, sizeof(pd[0]), o.ord);
        just_save_d(o, pd, ndir);
        continue;
      }
    }
  }


  if(verb>=2)
  {
    cout << "at the end nc_cap=" << nc_cap << ", nd_cap=" << nd_cap << endl;

    cout << "at the end nadc_cap=" << nadc_cap 
         << ", nad_cap=" << nad_cap 
         <<  ", id=" << id << ", nac=" << nac << endl;

    cout << "at the end ncf_cap=" << ncf_cap 
         << ", ndf_cap=" << ndf_cap
         <<  ", mf=" << mf << ", ncf=" << ncf << endl;
  }

  return 0;
}
