#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

int q=0;
char **gargv;
int gargc;

void usage(const char z[])
{
  if(q==0)
  {
    printf("USAGE: %s directory_name\n", gargv[0]);
    printf("  -q for no text output\n");
    printf("  -qq for limited text output\n");
    printf("  if directory_name exists and is not in the path then 1 is returned\n");
    printf("  0 is returned if the directory exists and is in the path\n");
    printf("  0 is returned if the directory does not exist\n");
    printf("USAGE: %s -f=file_name\n", gargv[0]);
    printf("  0 is returned if the file is not in the path\n");
    printf("  1 is returned if the file is in the path\n");
    printf("..stopping because: %s\n", z);
  }
  exit(0);
}

int compare(const char a[], const char b[])
{
  // .. case independent compare of a[] and b[]
  // .. with 0 returned if they are the same
  int i, la, lb;
  char ca, cb;
  la=strlen(a);
  lb=strlen(b);
  if(la!=lb)return 1;

  for(i=0; i<la; i++)
  {
    ca=a[i];
    if('A'<=ca && ca<='Z')ca='a'+ca-'A';
    cb=b[i];
    if('A'<=cb && cb<='Z')cb='a'+cb-'A';
    if(ca!=cb)return 1;
  }
  return 0;
}

int exist_file(const char d[], const char f[])
{
  // .. return 1 if file f[] is in directory d[],
  // .. return 0 if the file cannot be opened
  // .. if d[] is empty the f[] is assumed to be in the current directory
  // .. if f[] is "nul" then this checks for the existence of the
  // .. directory d[] on windows

  int i, j, ld, lf, k;
  char *z;
  FILE *fp;

  // .. create the full name as z[]
  ld=strlen(d);
  lf=strlen(f);
  z=(char*)malloc( (ld+lf+3)*sizeof(char) );
  j=0;
  for(i=0; i<ld; i++)
  {
    z[j]=d[i];
    j++;
  }
  if(j>0)
  {
    z[j]='\\';
    j++;
  }

  // .. exclude if only nul is given and no folder
  if(j==0 && strcmp(f, "nul")==0)
  {
    free(z);
    return 0;
  }

  for(i=0; i<lf; i++)
  {
    z[j]=f[i];
    j++;
  }
  z[j]=0;
  j++;

  // .. check if z[] can be read
  if(q==0 && j>1)printf("..checking existence of %s\n", z);

  k=0;
  fp=fopen(z, "rb");
  if(fp!=NULL)
  {
    k=1;
    fclose(fp);
  }

  free(z);
  return k;
}

void trim(char a[])
{
  int i, l;
  l=strlen(a);
  for(i=l-1; i>=0; i--)
    if(isspace(a[i]))a[i]=0;
    else             break;
}

int main(int argc, char *argv[])
{
  char *p, *c, *a, **d, *fulld, *fn, blank[1];
  int i, j, k, l, lf, m, n, lasti, lfn, jf; 
  blank[0]=0;

  gargc=argc;
  gargv=argv;

  if(argc<2)usage("no arguments supplied");
  q=0;
  for(i=1; i<argc; i++)
  {
    if(strcmp(argv[i], "-q")==0)q=1;
    if(strcmp(argv[i], "-qq")==0)q=2;
  }

  p=getenv("PATH");
  if(p==NULL)usage("no env PATH");
  trim(p);
  l=strlen(p);
  if(q==0)
  {
    printf("%s\n", p);
    printf("length l=%d\n", l);
  }

  // .. count the number of chars ; in p
  n=0;
  if(l>0)n=1;
  if(p[l-1]==';')l=l-1;
  for(i=0; i<l; i++)
    if(p[i]==';')n++;

  if(q==0)
    printf("number of folders in the path=%d\n", n);

  // .. get the folders as c-strings d[]
  c=(char*)malloc(sizeof(char)*l+1);
  d=(char**)malloc(sizeof(char*)*n+1);

  for(i=0; i<l; i++)
    c[i]=p[i];
  c[l]=0;

  d[0]=c;
  m=1;
  for(i=0; i<l; i++)
  {
    if(c[i]==';')
    {
      d[m]=&c[i+1];
      c[i]=0;
      m++;
    }
  }

  // .. make sure the c-strings are trimmed
  for(i=0; i<n; i++)
  {
    a=d[i];
    l=strlen(a);
    for(j=l-1; j>=0; j--)
      if(isspace(a[j])){a[j]=0; l=j;}
      else             break;
    for(j=0; j<l; j++)
    {
      if(isspace(a[j]))continue;
      d[i]=&a[j];
      break;
    }
  }

  // .. show all the folders
  for(i=0; i<n && q==0; i++)
    printf("%s\n", d[i]);

  // .. put all the argv[i] entries together to create one c-string
  lf=0;
  lasti=1;
  for(i=1; i<argc; i++)
  {
    a=argv[i];
    if(a[0]=='-')continue;
    l=strlen(a);
    lf=lf+l+1;
    lasti=i;
  }
  fulld=(char*)malloc(sizeof(char)*lf);
  fulld[0]=0;

  k=0;
  for(i=1; i<argc; i++)
  {
    a=argv[i];
    if(a[0]=='-')continue;
    l=strlen(a);
    for(j=0; j<l; j++)
    {
      fulld[k]=a[j];
      k++;
    }
    if(i<lasti)fulld[k]=' ';
    else       fulld[k]=0;
    k++;
  }

  if(k>0 && q==0)
    printf("fulld=%s\n", fulld);

  j=1;
  for(i=0; i<n && k>0; i++)
  {
    j=compare(fulld, d[i]);
    if(j==0)break;
  }

  if(j==1)
  {
    if(q==0 && k>0) printf("%s is not in the path\n", fulld);
    if(exist_file(fulld, "nul")==0)
    {
      if(q==0 && k>0) printf("folder %s does not exist\n", fulld);
      j=0;
    }
    if(q==0 && k>0) printf("folder %s does exist\n", fulld);
  }
  else
  {
    if(q==0 && k>0)
      printf("%s is in the path\n", fulld);
  }

  // .. check if a file is to be checked
  fn=blank;
  for(i=1; i<argc; i++)
  {
    a=argv[i];
    l=strlen(a);
    if(l<4)continue;
    if(strncmp(argv[i], "-f=", 3)!=0)continue;
    fn=&a[3];
    break;
  }
  lfn=strlen(fn);

  jf=0;
  if(exist_file("", fn)==1)
  {
    if(q==0 || q==2)printf("%s\n", fn);
    jf=2;
  }
  for(i=0; i<n && lfn>0 && jf==0; i++)
  {
    if(exist_file(d[i], fn)==1)
    {
      if(q==0 || q==2)printf("%s\\%s\n", d[i], fn);
      jf=2;
      break;
    } 
  }

  if(jf==0 && lfn>0 && (q==0 || q==2))
    printf("File %s   is_not_in_the_path\n", fn);

  free(c);
  free(d);
  free(fulld);

  if(lfn==0)
  {
    if(q==0)printf("..about to return %d to the operating system\n", j);
    return j;
  }
  else
  {
    if(q==0)printf("..about to return %d to the operating system\n", jf);
    return jf;
  }
}
