function print2pdf(fn)

%% printf2pdf(fn)
%    creates a pdf file called fn of the current graphics
%    with the size fitted to the size of the graphics instead of full page.

if nargin<1
  fprintf('Warning: No argument to %s given.\n', mfilename);
  fprintf('The argument should be the name of the pdf to create.\n');
  fprintf('Leaving the function, nothing created.\n');
  return;
end

% create ufn as the name to use
% and add .pdf if this is not already at the end
l=length(fn);
ufn=fn;
if l<5
  ufn=[fn(:)' '.pdf'];
  l=length(ufn);
end
if strcmp(ufn(l-3:l), '.pdf')==0
  ufn=[fn(:)' '.pdf'];
  l=length(ufn);
end
for i=1:l-4
  if(ufn(i)==' ' || ufn(i)=='.')
    ufn(i)='_';
  end
end

% remove any exist file called ufn
if exist(ufn, 'file')==2
  delete(ufn);
end

% save the initial values of the things that will be changed
prePaperPosition = get(gcf, 'PaperPosition');
prePaperSize = get(gcf, 'PaperSize');
p = prePaperPosition;

% Set the PaperPosition and PaperSize
% The first 2 entries of p indicate where it would have appeared 
% The plot itself is the same size as the -depsc, epstopdf mode but there
% is more surrounding whitespace with this version.

w = p(3);
h = p(4);

set(gcf,...
    'PaperPosition', [0 0 w h], ...
    'PaperSize',[w h]);

% create the pdf file
fprintf('..attempting to create %s\n', ufn);
print(ufn, '-dpdf', '-r600');

% restore the original values
set(gcf, 'PaperPosition', prePaperPosition);
set(gcf, 'PaperSize', prePaperSize);
