function sav_pref(bound)
%%function sav_pref(bound)
%%  Use as sav_pref() to create a zip file of the files in prefdir
%   below a default size of 100000.  If bound is precified then this
%   is the max file size used.

% get the preference folder
d=dir(prefdir);
n=length(d);

% get the end part of the name to be taken as the version
l=length(prefdir);
v=prefdir;
fprintf('prefdir=%s\n', v);
for k=l:-1:1
    x=v(k);
    if x=='\' || x=='/'
        v=v(k+1:l);
        break;
    end
end

% only save files below the size of bound
if nargin<1
  bound=100000;
end

% first count the number as m
m=0;
for k=1:n
    a=d(k).name;
    if a(1)=='.' || d(k).bytes>bound
        continue;
    end
    m=m+1;
end

% allocate space and get a cell array as z and the
% last modified times as lm
z=cell(m, 1);
lm=zeros(m, 1);
m=0;
for k=1:n
    a=d(k).name;
    if a(1)=='.'|| d(k).bytes>bound
        continue;
    end
    m=m+1;
    z{m}=[prefdir, '\\', a];
    lm(m)=get_from_date(d(k).date);
end
fprintf('%d files to backup\n');

[flm, ii]=max(lm);
f=z{ii};
fdet=dir(f);
fpart=fdet.date;
l=length(fpart);
for k=1:l
    if fpart(k)==' ' || fpart(k)==',' || fpart(k)==';' ...
      || fpart(k)=='.' || fpart(k)=='-' || fpart(k)==':'
      fpart(k)='_';
    end
end
fprintf('..the most recently changed file=\n%s\n', f);
fprintf('..the file time stamp=%s\n', fpart);


% get the current time to be used in file naming
c=clock;
s=floor(c(6));
ts=sprintf('%4d_%2d_%2d_%2d_%2d_%2d', c(1:5), s);
l=length(ts);
p=find(ts==' ');
ts(p)='0';
fprintf('..the current time=%s\n', ts);

% create a folder on h: to save
dn='h:\\mat_pref_save';
ie=exist(dn, 'dir');
if ie~=7
    fprintf('..attempting to create folder %s\n', dn);
    mkdir(dn);
else
    fprintf('..folder %s already exists\n', dn);
end

% create the name of the zip files
zn_lat=[dn, '\\', v, '_pref_latest.zip'];
zn_time=[dn, '\\', v, '_pref_', fpart, '.zip'];
fprintf('..saving as\n%s\nand\n%s\n', zn_time, zn_lat);

% remove any existing latest version and then create
ie=exist(zn_lat, 'file');
if ie==2
    delete(zn_lat);
end
zip(zn_lat, z);
zip(zn_time, z);

end


function n=get_from_date(a)
  % Get the number of seconds since 2000 using a date string.
  % This is needed when the datenum field is missing as in freemat.
  % It is a bit cumbersome as the different versions have different
  % formats for the date.
  
  l=length(a);
  y=0;
  mo=0;
  d=0;
  h=0;
  m=0;
  s=0;
  n=-1;
  
  % get the year
  for i=1:l-3
    u='0'<=a(i:i+3);
    v=a(i:i+3)<='9';
    if sum(u)<4 || sum(v)<4, continue, end;
    y=str2num(a(i:i+3))-2000;
    a(i:i+3)='    ';
    break;
  end
  
  % get the month number
  mons=['Jan'; 'Feb'; 'Mar'; 'Apr'; 'May'; 'Jun'; ...
        'Jul'; 'Aug'; 'Sep'; 'Oct'; 'Nov'; 'Dec'];
  for i=1:l-2
    if a(i)==' ', continue, end
    if '0'<=a(i) && a(i)<='9', continue, end
    for j=1:12
      if mons(j, :)==a(i:i+2)
        mo=j;
        a(i:i+2)='   ';
        break;
      end
    end
  end
  
  % get the hour, min and sec
  k=find(a==':', 1);
  if length(k)==0
    return;
  end
  if k<2
    return;
  end
  
  h=str2num(a(k-2:k-1));
  m=str2num(a(k+1:k+2));
  s=str2num(a(k+4:k+5));
  a(k-2:k+5)='        ';
  
  % remove anything which is not a digit that remains
  for i=1:l
    if a(i)<'0' || a(i)>'9'
      a(i)=' ';
    end
  end
  
  % any integers left should be the day of the month
  d=str2num(a);
  
  % now get the number of seconds since the start of the century
  day_sec=24*3600;
  n=s+60*m+3600*h;
  n=n+(d-1)*day_sec;
  md=[31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31];
  md4=md;
  md4(2)=md(2)+1;
  if rem(y, 4)==0
    mdu=md4;
  else
    mdu=md;
  end
  n=n+sum(mdu(1:mo-1))*day_sec;
  y4=floor(y/4);
  n=n+(365*y+y4)*day_sec;
end
