#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <cctype>
#include <cstring>
#include <climits>
#include <string>
#include <cmath>
#include <sys/stat.h>
#include <sys/types.h>

using namespace std;

// .. ensure that S_ISDIR macro is set for deciding if a directory
#ifndef S_ISDIR
#define S_ISDIR(m) ((m & S_IFMT) == S_IFDIR)
#endif

#if defined(__MINGW32__)
#ifndef WINVER
#define WINVER 0x0500
#endif
#endif

// ..macros to deal with pipes, processes and accurate timing 
// ..which are platform dependent


#if( defined(_WIN32) || defined(_WIN64) )
  #include <windows.h>
  #define POPEN _popen
  #define PCLOSE _pclose
  #ifndef _MSC_VER
    #include <unistd.h>
    #include <sys/time.h>
    #define GETPID getpid
  #else
    #include <process.h>
    #define GETPID _getpid
  #endif
#else
  #include <unistd.h>
  #include <sys/time.h>
  #define GETPID getpid
  #define POPEN popen
  #define PCLOSE pclose
#endif

#ifdef _MSC_VER
  #include <sys/utime.h>
#else
  #include <utime.h>
#endif

#if (_MSC_VER >= 1400)
  #define LONG long long
#else
  #define LONG long
#endif

// .. useful output or debugging macros
#define LINEPR(v) {cout << __LINE__ << ": " << #v << "=" << (v) << endl;}
#define PR(v) {cout << #v << "=" << (v) << endl;}
#define FPR(v) {if(quiet==0)cout << #v << "=" << (v) << endl; fout << #v << "=" << (v) << endl;}


class mkw_sys_prop {
public:
  int unx;        // .. set to 1 if appears to be the operating system
  int win;         // .. set to 1 if it appears to be windows
  int compiler64; // .. set to 1 if 64 bit pointers used by compiler
  int os64;       // .. set to 1 if 64 bit OS detected
  char *cpu;      // .. pointer to cpu indicated
  char os[12];    // .. operating system name
  char pathvs;    // .. ; or : used to separate parts of a path
  char nul[16];   // .. null device for unwanted output, /dev/null or nul
  char tmp[256];  // .. temporary location with write access
  char ipc[64];   // .. ipconfig or /sbin/ifconfig, commands to get ip address
  char kill[64];  // .. kill -9 or taskkill /f /pid
  char ps[16];    // .. ps -ax  or  tasklist /v
  char ip[64];    // .. ip address of pc
  char set[64];   // .. /bin/tcsh -c setenv    or just    set
  char mount[64]; // .. mount  or  net use
  char subst[64]; // .. subst (windows only)
  char s;         // .. / or \ for separating components of a file name
  int at_brunel;  // .. 1 if ip address starts with 134.83.
  char tmp_dr;    // .. drive letter actually used for tmp area
  char h_dr;      // .. drive letter actually used for h: drive (homedrive)
  char u_dr;      // .. drive letter actually used for u: drive
  char r_dr;      // .. drive letter actually used for r: drive
  char q_dr;      // .. drive letter actually used for local r: drive at brunel
  char t_dr;      // .. drive letter actually used for t: drive
  char l_dr;      // .. drive letter actually used for l: drive
  char *homeshare;// .. env variable which should agree with h: setting at brun
  char *username; // .. username env variable
  char *brunel_user; // .. brunel username
  char grp[64];   // .. group needed to mount h: drive
  unsigned int uid;      // .. the uid variable
  unsigned int gid;      // .. the gid variable
  char server[64];// .. server for u: drive
  char *userprofile; // .. userprofile env variable
  char *home;     // .. home env variable
  string exename; // .. excutable name of this program
  int n_pathext;    // .. number of components in the pathext variable
  char *c_pathext;
  char **cl_pathext;
  char pwd[512];  // .. current working directory
  int vpn;        // .. 1 if vpn running, 0 if not running
  string gvim;    // .. full name of a gvim on windows
  mkw_sys_prop();
  void show();
};

struct drives_prop {
  int available;       // .. 1 if a letter in not in use or pre-assigned
  char *description;   // .. as found using vbscript
  double free_space;   // .. free space as reported by vbscript
  char *network;       // .. blank or C-string giving details for net use
  char *subst;         // .. blank or the directory assigned to letter
  double size;         // .. reported capacity
  int attempt;         // .. 1 if attempt to assign so that only done once
  string unx;          // .. name as mounted on a linux pc
};


// .. class for reading textfiles and releasing the space later
class t_file {
public:
  char name[512];
  char *c;
  char **cl;
  int n;
  int nl;
  t_file(void);
  t_file(const char s[]);
  t_file(const char s[], int);
  ~t_file();
  void open(const char s[]);
  void open(const char s[], int i);
  void open(char *d, int m, char **dl, int ml);
  void show();
  void close();
};



class file_and_cout{
  private:
    ofstream f;
    string name;
  public:
    file_and_cout(const char s[]);
    ~file_and_cout();
    file_and_cout& operator<<(std::ostream &(*T)(std::ostream &));
    file_and_cout& operator<<(const char *a);
    file_and_cout& operator<<(int a);
    file_and_cout& operator<<(string& a);
    template <class T>
    file_and_cout& operator<<(const T& a);
};

file_and_cout::file_and_cout(const char s[])
{
  name=s;
  f.open(s);
  if(f)return;
  cout << ".. could not connect file name=" << s << " for output!!!" << endl;
  exit(0);
}

file_and_cout::~file_and_cout()
{
  if(!f.is_open())return;
  f.close();
}

file_and_cout& file_and_cout::operator<<(std::ostream &(*T)(std::ostream &))
{
  f << T;
  cout << T;
  return *this;
}

template <class T>
file_and_cout& file_and_cout::operator<<(const T& a)
{
  f << a;
  cout << a;
  return *this;
}
   
file_and_cout& file_and_cout::operator<<(const char *a)
{
  f << a;
  cout << a;
  return *this;
}

file_and_cout& file_and_cout::operator<<(int a)
{
  f << a;
  cout << a;
  return *this;
}

file_and_cout& file_and_cout::operator<<(string& a)
{
  f << a;
  cout << a;
  return *this;
}


char blank[1]="";
int quiet=0;            // set to 1 if to run in quiet mode, otherwise verbose
time_t g_now;
char time_now[15]="";
char time_now_20[21]="";
mkw_sys_prop sys_prop;
drives_prop dr_prop[26];    // .. info for all 26 letters
int gargc=0;
int pid=-1;            // ..process id when set
char **gargv;



void accurate_time(int& y, int& mon, int& d, int& h, int& m, int& s, int& ms);
void accurate_time(char c20[]);
void a2f(char c[], int n, const char s[]);
void command_output(const char a[], char o, t_file& t);
void command_output_nopipe(const char a[], char o, const char fn[], t_file& t);
void copy_t_file(t_file& t_to, t_file& t_from);
void cr_temp_name(char s[], const char a[]);
void ctime_c14_now(char c[], time_t& g_now);
int effective_strlen(const char a[]);
int exists(const char s[]);
int exists_dir(const char dname[]);
char* f2a(const char s[], int& n);
void file_details(const char fname[], time_t& age);
void file_details(const char fname[], char c14age[], int& bytes, int& isdir);
int filesize(const char s[]);
void get_cwd(char p[]);
void get_ip(t_file& t);
void get_list_name(string& list_name);
void get_list_root_pos(string& list_root_pos);
void get_other_env(t_file& t, t_file& tm);
void get_temp(t_file& t);
void getdir(string& p, string& sdir, const char lp[],
            const char let[], const char a[], int z);
int is_an_arg(const char a[], int& j);
int is_an_arg(const char a[]);
double large_file_size(const char fn[]);
void line_structure(char c[], int n, char**& cl, int& nl);
void makelower(char c[]);
void makeupper(char c[]);
int mkwfind(const char c[], const char s[]);
int mkwfind_nocase(const char c[], const char s[]);
void mkw_ls(const char d[], const char rec[], int& n, char**& fn, t_file& t);
int mkw_utime(const char fname[], time_t t);
void mkwsystem(const char s[]);
void mstrcpy(char a[], const char b[]);
void mstrncpy(char a[], const char b[], const int n);
void p2a(const char a[], char*& c, int& n);
void replace(string& s, char a, char b);
void replace_string(char c[], char d[], int& m, const char a[], const char b[]);
void set_cwd_windows();
void textfile(const char s[], char*& c, int& n, int& nl, char**& cl);
void textfile(const char s[], char*& c, int& nl, char**& cl);
void textfile2(const char s[], char*& c, int& n, int& nl, char**& cl);
void time_t_to_struct_tm(time_t& g, struct tm& t);
void unique_sort(const void *a, int n, int sz, 
                 int(*f)(const void*, const void*), int o[], int& m, int g[]);

mkw_sys_prop::mkw_sys_prop()
{
  // .. settings until decisions are made
  unx=0;
  win=0;
  mstrcpy(nul, "");
  mstrcpy(ipc, "");

  // .. initialise letters used for certain drives before they are set
  tmp_dr=0;
  h_dr=0;
  u_dr=0;
  q_dr=0;
  r_dr=0;
  t_dr=0;
  l_dr=0;
  brunel_user=blank;
  uid=0;
  gid=0;

  // .. decide on the operating system
  if (exists("/etc/csh.cshrc"))unx=1;
  else
  {
    if (exists("c:\\windows\\system32\\ipconfig.exe"))
    {
      win=1;
    }
  }
  if(unx)gvim="gvim";
  else gvim="notepad";

  // .. set things dependent on whether unix or windows
  // .. that can be done easily
  exename="";
  if (unx)
  {
    s='/';
    mstrcpy(nul, "/dev/null");
    mstrcpy(ipc, "/sbin/ifconfig");
    mstrcpy(set, "/bin/tcsh -c setenv");
    mstrcpy(mount, "mount");
    mstrcpy(subst, "");
    mstrcpy(os, "unix");
    mstrcpy(kill, "kill -9 ");
    mstrcpy(ps, "ps -ax ");
    pathvs=':';
  }
  else
  {
    s='\\';
    mstrcpy(nul, "nul");
    mstrcpy(ipc, "ipconfig.exe");
    mstrcpy(set, "set");
    mstrcpy(mount, "net use");
    mstrcpy(subst, "subst");
    if (win==1)   mstrcpy(os, "win");
    mstrcpy(kill, "taskkill /f /pid ");
    mstrcpy(ps, "tasklist /v ");
    pathvs=';';
  }
}

void mkw_sys_prop::show()
{
  cout << "cpu=" << cpu << endl;
  if (compiler64==1) cout << "64 bit compiler" << endl;
  if (os64==1) cout << "64 bit os" << endl;
  if (unx)  cout << "Unix type operating system" << endl;
  if (win)   cout << "Windows operating system" << endl;
  if (unx==0 && win==0)
    cout << "no operating system identified yet" << endl;
  if (strlen(ipc)>0)cout << "To get ip address: " << ipc << endl;
  else             cout << "Do not know how to get ip address" << endl;
  if (strlen(ip)>0)cout << "IP address is " << ip << endl;
  else            cout << "No IP address identified" << endl;
  if (strlen(tmp)>0)cout << "Temporary directory=" << tmp << endl;
  else             cout << "no temporary directory set yet" << endl;
  if (strlen(nul)>0)cout << "Null location is " << nul << endl;
  else             cout << "No null location set yet" << endl;

  cout << "username=" << username << endl;
  cout << "brunel_user=" << brunel_user << endl;
  cout << "group=" << grp << endl;
  if (at_brunel && unx==0)
  {
    if(strlen(homeshare)>0)
      cout << "homeshare variable needed for h: drive=" << homeshare << endl;
    else
    {
      cout << "homeshare variable not set, possibly a vpn connection" << endl;
      if(vpn==1)
      {
        cout << "vpnclient indicates that there is a vpn connection." << endl;
        cout << "Mount drives  L:, H:, R:, U: outside of this program." 
             << endl;
      }
    }

    cout << "server for u: is probably=" << server << endl;
  }
  else
  {
    if (at_brunel==0)
      cout << "IP address suggest not on the Brunel network" << endl;
  }
  if (pwd!=0)
    cout << "Current working directory is " << pwd << endl;
}

void t_file::close()
{
  if (n>0)delete[] c;
  if (nl>0)delete[] cl;
  c=NULL;
  cl=NULL;
  n=-1;
  nl=-1;
}

void t_file::open(char *d, int m, char **dl, int ml)
{
  c=d;
  cl=dl;
  n=m;
  nl=ml;
  mstrcpy(name, "possibly_pipe_output");
}

void t_file::open(const char s[])
{
  char *d, **dl;
  int m, ml;
  textfile(s, d, m, ml, dl);
  c=d;
  cl=dl;
  n=m;
  nl=ml;
  mstrcpy(name, s);
}

void t_file::open(const char s[], int)
{
  char *d, **dl;
  int m, ml;
  textfile2(s, d, m, ml, dl);
  c=d;
  cl=dl;
  n=m;
  nl=ml;
  mstrcpy(name, s);
}

void t_file::show()
{
  int i;
  for (i=0; i<nl; i++)cout << cl[i] << endl;
}

t_file::t_file(void)
{
  name[0]=0;
  c=NULL;
  cl=NULL;
  n=-1;
  nl=-1;
}

t_file::t_file(const char s[])
{
  char *d, **dl;
  int m, ml;
  textfile(s, d, m, ml, dl);
  c=d;
  cl=dl;
  n=m;
  nl=ml;
  mstrcpy(name, s);
}

t_file::t_file(const char s[], int)
{
  // .. the 2nd arg is just to make it verbose
  char *d, **dl;
  int m, ml;
  textfile2(s, d, m, ml, dl);
  c=d;
  cl=dl;
  n=m;
  nl=ml;
  mstrcpy(name, s);
}

t_file::~t_file()
{
  if (n>0)delete[] c;
  if (nl>0)delete[] cl;
  c=NULL;
  cl=NULL;
  n=-1;
  nl=-1;
}


void accurate_time(int& y, int& mon, int& d, int& h, int& m, int& s, int& ms)
{
  ms=0;
  #if ( ( ( defined(_WIN32) || defined(_WIN64) ) && !defined(_MSC_VER) ) || ( !(defined(_WIN32)) && !(defined(_WIN64)) ) )
    time_t t;
    struct tm tm;

    #if ( (GCC_VERSION > 3042) || (!(defined(_WIN32)) && !(defined(_WIN64))) )
      struct timezone tz;
      struct timeval tv;
      gettimeofday(&tv, &tz);
      t=tv.tv_sec;
      ms=tv.tv_usec;  //..microseconds
    #else
      time(&t);
    #endif
    time_t_to_struct_tm(t, tm);

    y=1900+tm.tm_year;
    mon=1+tm.tm_mon;
    d=tm.tm_mday;
    h=tm.tm_hour;
    m=tm.tm_min;
    s=tm.tm_sec; 
  #else
    SYSTEMTIME st;
    //GetSystemTime(&st);
    GetLocalTime(&st);
    y=st.wYear;
    mon=st.wMonth;
    d=st.wDay;
    h=st.wHour;
    m=st.wMinute;
    s=st.wSecond;
    ms=1000*st.wMilliseconds;  //..microseconds
  #endif
}

void accurate_time(char c20[])
{
  int y, mon, d, h, m, s, ms;
  accurate_time(y, mon, d, h, m, s, ms);
  stringstream st;
  st.clear();
  st.str("");
  st << setfill('0');
  st << setw(4) << y << setw(2) << mon << setw(2) << d
     << setw(2) << h << setw(2) << m << setw(2) << s << setw(6) << ms;
  string z=st.str();
  mstrncpy(c20, z.c_str(), 20);
  c20[20]=0;
}

void a2f(char c[], int n, const char s[])
{
  // .. write c[] to file with name s[]
  if (n<0)return;

  ofstream inp(s, ios::binary);
  if (n>0)inp.write(c, n);
  inp.close();
}

void bit_64_question(t_file& tenv)
{
  // .. attempt to determine if the compiler and/or OS is 64 bit
  // .. use pointers to deduce the compiler being used
  int i, j;
  i=sizeof(char*);
  sys_prop.compiler64=0;
  if(i==8) sys_prop.compiler64=1;
  sys_prop.os64=-1;
  sys_prop.cpu=blank;

  char *d;
  if(sys_prop.unx)
  {
    for(i=0; i<tenv.nl; i++)
    {
      d=tenv.cl[i];
      if(mkwfind_nocase(d, "CPU=")==0)
      {
        sys_prop.cpu=&d[4];
        break;
      }
    }
  }
  else
  {
    for(i=0; i<tenv.nl; i++)
    {
      d=tenv.cl[i];
      if(mkwfind_nocase(d, "PROCESSOR_ARCHITECTURE=")==0)
      {
        j=mkwfind(d, "=");
        sys_prop.cpu=&d[j+1];
      }
      if(mkwfind_nocase(d, "PROCESSOR_ARCHITEW6432=")==0)
      {
        j=mkwfind(d, "=");
        sys_prop.cpu=&d[j+1];
      }
    }
  }

  if(strlen(sys_prop.cpu)>0)
  {
    sys_prop.os64=0;
    if(mkwfind(sys_prop.cpu, "64")>=0)sys_prop.os64=1;
  }
}

void command_output(const char a[], char o, t_file& t)
{
  // .. a is the command, o is 'a' or 'o' or 'n' depending on whether
  // .. all output, just ordinary output or no output is needed
  // .. the output is collected in t
  // .. this version uses pipes as it may not be known get
  // .. as to where it is possible to write files

  // .. use another function which writes to files
  // .. if only the error output is needed

  string s;
  s=a;

  // .. error output as well case
  if (o=='a')s=s+" 2>&1";

  // .. just the ordinary output case with error output ignored
  if (o=='o')s=s+" 2> "+sys_prop.nul;

  if (o=='n')s=s+" >"+sys_prop.nul +" 2>&1";

  char *c, **cl;
  int n, nl;

  p2a(s.c_str(), c, n);

  line_structure(c, n, cl, nl);
  t.open(c, n, cl, nl);

}

void command_output_nopipe(const char a[], char o, const char fn[], t_file& t)
{
  // .. a is the command, o is 'a' or 'o' depending on whether
  // .. all output or just ordinary output is needed
  // .. the output is collected in t

  // .. set the output file name in tmp using fn[] and the date
  // .. the date is used to make it less likely that 2 processes are
  // .. are attempting to write to the same file at the same time
  string ou, ouq;
  ou=sys_prop.tmp;
  ou=ou+sys_prop.s+fn+time_now_20;
  ouq='"';
  ouq=ouq+ou+'"';

  string s;
  s=a;

  // .. error output as well case
  if (o=='a') s=s+"  >"+ouq+" 2>&1";

  // .. just the ordinary output case with error output ignored
  if (o=='o')s=s+"  >"+ouq+" 2> "+sys_prop.nul;

  // .. execute the command
  mkwsystem(s.c_str());

  // .. read the output file
  if (o=='n')t.nl=-1;
  else
    t.open(ou.c_str());

}

void copy_bytes(const char from[], const char to[])
{
  const int N=65536;
  char c[N];

  if(exists(from)==0)return;

  // .. open the files
  ifstream in(from, ios::binary);
  ofstream ou(to, ios::binary);

  // .. read in blocks and write in blocks
  LONG nr;
  do
  {
    in.read(c, N);
    nr=in.gcount();
    if(nr>0)ou.write(c, nr);
  }while(nr==N);
  in.clear();
  in.close();
  ou.close();
}

void copy_file(const char from[], const char to[])
{
  // .. copy file from[] to to[] preserving the timestamp
  copy_bytes(from, to);

  // .. get details of from[] and then set the timestamp of to[]
  time_t t;
  file_details(from, t);
  mkw_utime(to, t);
}

void copy_t_file(t_file& t_to, t_file& t_from)
{
  // .. allocate the space assoc with t_to to have the same contents as t_from
  int i, n, nl;
  n=t_from.n;
  nl=t_from.nl;

  string s;
  s=t_from.name;
  s=s+" : a copy";
  mstrcpy(t_to.name, s.c_str());

  // .. allocate the space and copy the chars
  t_to.n=n;
  t_to.nl=nl;
  t_to.c=new char[n+1];
  t_to.cl=new char*[nl+1];
  for (i=0; i<n; i++)t_to.c[i]=t_from.c[i];

  // .. now set the line structure to be the same as t_from
  t_to.cl[0]=t_to.c;
  for (i=1; i<nl; i++)
    t_to.cl[i]=t_to.cl[i-1]+(t_from.cl[i]-t_from.cl[i-1]);

}

void cr_temp_name(char s[], const char a[])
{
  // .. create s[] by combining the temporary location and a[]
  string stem;

  // .. use the drive letter if this available
  if (sys_prop.unx==0 && sys_prop.tmp_dr!=0)
  {
    stem=sys_prop.tmp_dr;
    stem=stem+":";
  }
  else
    stem=sys_prop.tmp;

  stem=stem+sys_prop.s+a;
  int i, l;
  l=stem.length();
  if (l>1023) {
    cout << stem << " has length>1023" << endl;
    exit(-1);
  }
  const char *p=stem.c_str();
  for (i=0; i<l; i++)s[i]=p[i];
  s[l]=0;
}

void ctime_c14_now(char c[], time_t& g_now)
{
  // .. create the current time as a 14 digits yyyymmddhhmmss
  time(&g_now);
  struct tm t;
  time_t_to_struct_tm(g_now, t);

  stringstream sstem;
  string stem;
  sstem.clear();
  sstem.str("");
  sstem << 1900+t.tm_year << setw(2) << 1+t.tm_mon << setw(2) << t.tm_mday
  << setw(2) << t.tm_hour << setw(2) << t.tm_min << setw(2) << t.tm_sec;
  stem=sstem.str();
  mstrcpy(c, stem.c_str());
  int i;
  for (i=0; i<14; i++)if (!isdigit(c[i]))c[i]='0';
}

int effective_strlen(const char a[])
{
  // .. get the length of a[] ignoring trailing blanks
  int i, l;
  l=strlen(a);
  for (i=l-1; i>=0; i--)
    if (isspace(a[i]))continue;
    else             return i+1;
  return 0;
}

int exists(const char s[])
{
  // .. return 1 if file s[] can be opened, return 0 otherwise
  // .. the existence of the file is classified using this
  ifstream i(s);
  if (i.is_open())
  {
    i.close();
    return 1;
  }
  return 0;
}

int exists_dir(const char dname[])
{
  // .. return 1 if dname[] is a directory
  struct stat finfo;
  unsigned short mty;
  int ij=stat(dname, &finfo);

  // .. deal with cases when the file is not accessed
  if (ij!=0) return 0;
  mty=finfo.st_mode;

  // .. indicate if a directory
  if (S_ISDIR(mty))return 1;
  return 0;
}

void other_file_prop(const char fn[], char rwx[])
{
  // .. set rwx[] depending on whether the user of the program
  // .. has read, write or execute permission on the file
  if(fn[0]==0)return;
  rwx[0]='-';
  rwx[1]='-';
  rwx[2]='-';
  rwx[3]=0;
  if(sys_prop.unx==0)return;

#if( !defined(_WIN32) && !defined(_WIN64) )
  struct stat finfo;
  unsigned short mty;
  int ij=stat(fn, &finfo);
  if(ij!=0)return;

  // .. get mode and ownership of the file 
  mty=finfo.st_mode;
  uid_t u_id=finfo.st_uid;
  gid_t g_id=finfo.st_gid;

  if( u_id==sys_prop.uid)
  {
    if(mty & S_IRUSR)rwx[0]='r';
    if(mty & S_IWUSR)rwx[1]='w';
    if(mty & S_IXUSR)rwx[2]='x';
  }
  else
  {
    if(g_id==sys_prop.gid)
    {
      if(mty & S_IRGRP)rwx[0]='r';
      if(mty & S_IWGRP)rwx[1]='w';
      if(mty & S_IXGRP)rwx[2]='x';
    }
    else
    {
      if(mty & S_IROTH)rwx[0]='r';
      if(mty & S_IWOTH)rwx[1]='w';
      if(mty & S_IXOTH)rwx[2]='x';
    }
  }
#endif
}

char* f2a(const char s[], int& n)
{
  // .. read the entire file into an array and return the pointer
  n=filesize(s);
  if (n<0)return NULL;

  // .. allocate 1 more space as a char may be added in textfile()
  char *c=new char[n+1];
  ifstream inp(s, ios::binary);
  inp.read(c, n);
  inp.close();
  return c;
}

void file_details(const char fname[], time_t& age)
{
  struct stat finfo;
  //unsigned short mty;
  int ij=stat(fname, &finfo);

  // .. deal with cases when the file is not accessed
  if (ij!=0)
  {
    age=0;
    return;
  }
  age=finfo.st_mtime;
  //mty=finfo.st_mode;

  // .. alter any dates out of range to prevent failure
  if (age<0)age=0;
}

void file_details(const char fname[], char c14age[], int& bytes, int& isdir)
{
  struct stat finfo;
  time_t age;
  unsigned short mty;
  isdir=0;
  int ij=stat(fname, &finfo);

  // .. deal with cases when the file is not accessed
  if (ij!=0)
  {
    mstrcpy(c14age, "00000000000000");
    bytes=0;
    return;
  }
  age=finfo.st_mtime;
  mty=finfo.st_mode;

  // .. alter any dates out of range to prevent failure
  if (age<0)age=0;

  // .. indictate if a directory
  if (S_ISDIR(mty))isdir=1;

  struct tm g;
  time_t_to_struct_tm(age, g);
  stringstream sstem;
  string stem;

  sstem.clear();
  sstem.str("");
  sstem << 1900+g.tm_year << setw(2) << 1+g.tm_mon << setw(2) << g.tm_mday
  << setw(2) << g.tm_hour << setw(2) << g.tm_min << setw(2) << g.tm_sec;
  stem=sstem.str();
  mstrcpy(c14age, stem.c_str());

  int i;
  for (i=0; i<14; i++)if (!isdigit(c14age[i]))c14age[i]='0';

  bytes=int(finfo.st_size);
}

int filesize(const char s[])
{
  // .. get the file size by reading as much as possible
  LONG nr, ns;
  int n;
  const int N=65536;
  char c[N];

  // .. open the file and read to get the size
  if (exists(s)==0)return -1;
  ifstream inp(s, ios::binary);
  ns=0;
  do
  {
    inp.read(c, N);
    nr=inp.gcount();
    ns=ns+nr;
  } while (nr==N);
  inp.clear();
  inp.close();
  n=int(ns);
  return n;
}

void get_cwd(char acwd[], int lcwd)
{
  #ifndef _MSC_VER
    getcwd(acwd, lcwd);
  #else
    GetCurrentDirectory(lcwd, acwd);
  #endif
  acwd[lcwd-1]=0;
}

void get_ip(t_file& t)
{
  // .. get the ip address
  // .. this seems to work for windows or linux
  int done, i, j, j2, js, k1, k2, k3, kc, l;
  char *d;
  done=0;
  for (i=0; i<t.nl; i++)
  {
    d=t.cl[i];
    k1=mkwfind(d, "IP Address");
    k2=mkwfind(d, "IPv4 Address");
    k3=mkwfind(d, "inet addr:");
    kc=mkwfind(d, ":");
    if (kc<0)continue;
    if (k1>0 || k2>0 || k3>0)
    {
      l=strlen(d);
      for (j=kc+1; j<l; j++)
        if (isdigit(d[j]))
        {
          js=l-1;
          for (j2=j+1; j2<l; j2++)if (isspace(d[j2])) {
              js=j2-1;
              break;
            }
          mstrncpy(sys_prop.ip, &d[j], js-j+1);
          sys_prop.ip[js-j+1]=0;
          done++;
          break;
        }
    }
    if (done>0 && mkwfind(sys_prop.ip, "134.83.")==0)break;
  }

  // .. determine if on the brunel network and possibly
  // .. reset the address
  sys_prop.at_brunel=0;
  for(i=0; i<t.nl; i++)
  {
    d=t.cl[i];
    l=strlen(d);
    k1=mkwfind(d, "134.83.");
    if(k1>=0)
    {
      int l1=strlen(&d[k1]);
      l=mkwfind(&d[k1], " ");
      if(l>0 && l<l1)l1=l;
      mstrncpy(sys_prop.ip, &d[k1], l);
      sys_prop.ip[l1]=0;
      sys_prop.at_brunel=1;
      break;
    }
  }
}

void get_lfn(const char a[], int lb, char b[])
{
  int k=0, l=strlen(a);
#if( defined(_WIN32) || defined(_WIN64) )
  k=GetLongPathName(a, b, lb-1);
#endif
  if(l>lb-1)l=lb-1;
  if(k==0)
  {
    mstrncpy(b, a, l);
    b[l]=0;
  }
}

void get_list_name(string& list_name)
{
  int i;
  list_name="";
  for (i=1; i<gargc-1; i++)
  {
    if (strcmp(gargv[i], "-list_name")==0)
    {
      if (exists(gargv[i+1]))
      {
        list_name=gargv[i+1];
        break;
      }
    }
  }
}

void get_list_root_pos(string& list_root_pos)
{
  int i;
  list_root_pos=sys_prop.pwd;
  for (i=1; i<gargc-1; i++)
  {
    if (strcmp(gargv[i], "-list_root_pos")==0 && gargv[i+1][0]!='-')
    {
      list_root_pos=gargv[i+1];
      break;
    }
  }
}


void get_other_env(t_file& t, t_file& tm)
{
  // .. get some of the other env variables needed
  sys_prop.homeshare=blank;
  sys_prop.username=blank;
  sys_prop.userprofile=blank;

  int i, j, k, l;
  char *d;
  for (i=0; i<t.nl; i++)
  {
    d=t.cl[i];
    j=mkwfind(d, "HOMESHARE=");
    if (j>=0)
    {
      k=mkwfind(d, "=");
      sys_prop.homeshare=&d[k+1];
    }

    j=mkwfind(d, "USERNAME=");
    if (j>=0)
    {
      k=mkwfind(d, "=");
      sys_prop.username=&d[k+1];
    }

    j=mkwfind(d, "USER=");
    if (j>=0 && sys_prop.unx==1)
    {
      k=mkwfind(d, "=");
      sys_prop.username=&d[k+1];
    }

    j=mkwfind(d, "GROUP=");
    if (j>=0 && sys_prop.unx==1)
    {
      k=mkwfind(d, "=");
      mstrcpy(sys_prop.grp, &d[k+1]);
    }

    j=mkwfind(d, "USERPROFILE=");
    if (j>=0)
    {
      k=mkwfind(d, "=");
      sys_prop.userprofile=&d[k+1];
    }

    j=mkwfind(d, "home=");
    if (j>=0)
    {
      k=mkwfind(d, "=");
      sys_prop.home=&d[k+1];
    }

    j=mkwfind(d, "PWD=");
    if (j>=0)
    {
      k=mkwfind(d, "=");
      mstrcpy(sys_prop.pwd, &d[k+1]);
    }


    j=mkwfind(d, "gid=");
    if (j>=0)
    {
      k=mkwfind(d, "=");
      sys_prop.gid=atoi(&d[k+1]);
    }


  }
  if (sys_prop.unx==0)
    get_cwd(sys_prop.pwd, 511);

  if(sys_prop.at_brunel)
  {
    sys_prop.brunel_user=sys_prop.username;

    // ..search the net use output
    int done=0;
    for(i=0; i<tm.nl; i++)
    {
      d=tm.cl[i];
      if(mkwfind(d, " H:")<0)continue;
      if(mkwfind(d, "134.83.")<0 && mkwfind(d, "acfs1\\")<0)continue;
      l=strlen(d);
      for(j=l-1; j>=0; j--)
        if(d[j]=='\\')
        {
          sys_prop.brunel_user=&d[j+1];
          done=1;
          for(k=j-1; k>=0; k--)
          {
            if(d[k]=='\\')
            {
              mstrncpy(sys_prop.grp, &d[k+1], j-k-1);
              sys_prop.grp[j-k-1]=0;
              break;
            }
          }
          break;
        }
      if(done)break;
    }

    // .. check env variables if still not set
    if(done==0)
    {
      for (i=0; i<t.nl; i++)
      {
        d=t.cl[i];
        j=mkwfind(d, "brunel_user=");
        if (j>=0)
        {
          k=mkwfind(d, "=");
          sys_prop.brunel_user=&d[k+1];
        }
        j=mkwfind(d, "brunel_grp=");
        if (j>=0)
        {
          k=mkwfind(d, "=");
          mstrcpy(sys_prop.grp, &d[k+1]);
        }
      }
    }
  }
  l=mkwfind(sys_prop.brunel_user, " ");
  if(l>0)sys_prop.brunel_user[l]=0;

  #if( !defined(_WIN32) && !defined(_WIN64) )
  sys_prop.uid=getuid();
  sys_prop.gid=getgid();
  #endif
}

void get_temp(t_file& t)
{
  // .. set the temporary directory
  // .. t here contains the output from setenv or set commands
  int i, k1, k2, ke;
  char *d;
  if (sys_prop.unx)
  {
    if(exists_dir("/home/latex/"))
      mstrcpy(sys_prop.tmp, "/home/latex");
    else
      mstrcpy(sys_prop.tmp, "/tmp");
  }
  else
  {
    for (i=0; i<t.nl; i++)
    {
      d=t.cl[i];
      k1=mkwfind_nocase(d, "temp=");
      k2=mkwfind_nocase(d, "tmp=");
      ke=mkwfind_nocase(d, "=");
      if (ke<0)continue;
      if (k1!=0 && k2!=0)continue;
      mstrcpy(sys_prop.tmp, &d[ke+1]);
      break;
    }
  }

  // .. test that we can write to the temporary directory
  if (strlen(sys_prop.tmp)==0)return;
  string s;
  s=sys_prop.tmp;
  s=s+sys_prop.s+"write_test";

  int n;
  const int N=4096;
  char w[N];
  for (i=0; i<N; i++)w[i]=0;
  a2f(w, N, s.c_str());
  n=filesize(s.c_str());
  if (n!=N)
  {
    cout << "WARNING: write test to " << s << " failed" << endl;
  }
  if (exists(s.c_str()))remove(s.c_str());
}

void getdir(string& p, string& sdir, const char lp[],
            const char let[], const char a[], int z)
{
  // .. s is a string containing many file names separated by ;
  // .. if z==1 then
  // .. sdir is determined corresponding to the first file that exists.
  // .. else
  // .. sdir is all the directories for which a file exists
  // .. lp[] is the last part in p which needs to be removed
  // .. a[] contains info relating to excluding things in the search.
  // .. if p contains %c: then all available drives in let[] are searched

  int llp=strlen(lp);

  // .. get the components of p in dl[], j is the number of components
  int i, id, j, l;
  char d[16384];
  char *dl[256];
  mstrcpy(d, p.c_str());
  l=strlen(d);
  dl[0]=d;
  j=0;
  for (i=1; i<l; i++)
  {
    if (d[i]==';')
    {
      d[i]=0;
      j++;
      dl[j]=&d[i+1];
    }
  }

  int lj=strlen(let);

  // .. now search the j components
  sdir="";
  int ndir=0;
  string s, sr;
  int ls, ij;


  for (i=0; i<j; i++)
  {
    if (strncmp(dl[i], "%c:", 3)==0)
    {
      for (ij=0; ij<lj; ij++)
      {
        id=let[ij]-'A';

        if(id<0 || id>25)continue;
        // . skip non-existent drives

        if (dr_prop[id].size<=0)continue;
        if (dr_prop[id].available==1)continue;
        if (dr_prop[id].size<=0)continue;

        // . possibly skip network drives depending in a[]
        if (strlen(dr_prop[id].network)>0 &&
            mkwfind(a, "not network")>=0) continue;

        // .. set the name as s
        s=let[ij];
        s=s+":"+&dl[i][3];
        replace(s, '/', '\\');

        if (exists(s.c_str())==0)continue;
        ndir++;
        ls=s.length();
        sr.assign(s, 0, ls-llp);
        if(z==1)
        {
          sdir=sr;
          return;
        }
        else
        {
          sdir=sdir+";"+sr+";";
        }
      }
    }
    else
    {
      s=dl[i];
      ls=s.length();
      if(sys_prop.unx==0)replace(s, '/', '\\');
      sr.assign(s, 0, ls-llp);
      if (exists(s.c_str()))
      {
        if(z==1)
        {
          sdir=sr;
          return;
        }
        else
        {
          sdir=sdir+";"+sr+";";
        }
      }
    }
  }
}

int is_an_arg(const char a[], int& j)
{
  // .. return 1 if a[] is one of the command line arguments
  int i;
  j=-1;
  for (i=1; i<gargc; i++)
    if (strcmp(a, &gargv[i][1])==0)
    {
      j=i;
      return 1;
    }
  return 0;
}

int is_an_arg(const char a[])
{
  int j;
  return is_an_arg(a, j);
}

double large_file_size(const char fn[])
{
  LONG bytes;
  double dbytes;

  #ifdef _MSC_VER
    //..cl.exe version after version 6
    #if (_MSC_VER >= 1400)
      unsigned int a[2];
      HANDLE H=CreateFile(fn, GENERIC_READ, FILE_SHARE_READ, NULL, 
                        OPEN_EXISTING, SECURITY_ANONYMOUS, NULL); 
      GetFileSizeEx(H, (LARGE_INTEGER*)a);
      dbytes=a[0];
      if(a[1]>0)dbytes=dbytes+a[1]*2*2147483648.0;
      bytes=(LONG)floor(dbytes);
      CloseHandle(H);
    #else
      struct stat finfo;
      int ij=stat(fn, &finfo);
      if (ij!=0) bytes=0;
      else bytes=finfo.st_size;
      dbytes=bytes;
    #endif
  #else
    #if( defined(_WIN32) || defined(_WIN64) )
      //..mingw version on windows which are new enough
      #if GCC_VERSION > 3023
        __stat64 statbf;
        uint64_t size64;
        _stat64(fn, &statbf);
        size64=statbf.st_size;
        dbytes=size64;
      #else
        struct stat finfo;
        int ij=stat(fn, &finfo);
        if (ij!=0) bytes=0;
        else bytes=finfo.st_size;
        dbytes=bytes;
        
      #endif
    #else
      // ..linux version
      struct stat64 finfo;

      int ij=stat64(fn, &finfo);

      // .. deal with cases when the file is not accessed
      if (ij!=0) bytes=0;
      else bytes=finfo.st_size;

      dbytes=bytes;
    #endif
  #endif
  return dbytes;
}

void line_structure(char c[], int n, char**& cl, int& nl)
{

  // .. gets the line structure associated with c[] of length n

  // .. count the lines by counting the number of 10s and 13s
  nl=0;
  int i, nl10=0, nl13=0, ieol=10;
  for (i=0; i<n; i++)
  {
    if (c[i]==10)nl10++;
    if (c[i]==13)nl13++;
  }

  // .. usually use 10 but use 13 if possibly a mac file
  nl=nl10;
  ieol=10;
  if (1<=nl13 && 4*nl10<=nl13)
  {
    nl=nl13;
    ieol=13;
  }
  // .. set the lines and make them null terminated
  cl=new char*[nl+1];

  // .. but do not do anything if n<=0 and take case of c[]
  if (n<0)c=new char[1];
  if (n<=0) return;

  // .. first line
  cl[0]=c;
  nl=1;

  // .. subsequent lines
  for (i=0; i<n-1; i++)
    if (c[i]==ieol)
    {
      cl[nl]=&c[i+1];
      nl++;
    }

  // .. add a char if the last line is not completed
  if (c[n-1]!=ieol)
  {
    c[n]=0;
    n++;
  }

  // .. ensure that all lines are null-terminated
  for (i=0; i<n; i++)
  {
    if (c[i]==13)c[i]=0;
    if (c[i]==10)c[i]=0;
  }
}

void makelower(char c[])
{
  // .. convert c[] to lowercase
  int i, l;
  l=strlen(c);
  for (i=0; i<l; i++)
    if ('A'<= c[i] && c[i]<='Z')c[i]=char(c[i]+'a'-'A');
}

void makeupper(char c[])
{
  // .. convert c[] to uppercase
  int i, l;
  l=strlen(c);
  for (i=0; i<l; i++)
    if ('a'<= c[i] && c[i]<='z')c[i]=char(c[i]+'A'-'a');
}

int mkwfind(const char c[], const char s[])
{
  // .. locate the 1st position of s[] in c[], -9 if no match
  int ls=strlen(s);
  int lc=strlen(c);
  if (ls>lc)return -9;
  int i, j, m;
  for (i=0; i<=lc-ls; i++)
  {
    m=0;
    for (j=0; j<ls; j++)
    {
      if (c[i+j]==s[j])m++;
      else break;
    }
    if (m<ls)continue;
    return i;
  }
  return -9;
}

int mkwfind_nocase(const char c[], const char s[])
{
  // .. locate the 1st position of s[] in c[] ignoring case, -9 if no match
  int ls=strlen(s);
  int lc=strlen(c);
  if (ls>lc)return -9;
  int i, j, m;
  for (i=0; i<=lc-ls; i++)
  {
    m=0;
    for (j=0; j<ls; j++)
    {
      if ( tolower(c[i+j])==tolower(s[j]) )m++;
      else break;
    }
    if (m<ls)continue;
    return i;
  }
  return -9;
}

void mkw_ls(const char d[], const char rec[], int& n, char**& fn, t_file& t)
{
  // ..d[] is the directory to list,
  // ..if rec[] is empty then only this directory, otherwise also sub-dir
  // ..on exit n=number of files
  // ..fn[0], ..., fn[n-1] point to the names
  // ..t is the textfile type containing what fn[] points to
  n=-1;
  fn=NULL;

  // ..first check that d[] is a directory
  if(exists_dir(d)==0)return;

  // ..create the command as s for the listing and the output name as so
  string s, so;
  if(strlen(rec)==0)
  {
    if(sys_prop.unx==1) s="ls ";
    else                s="dir /a/b ";
  }
  else
  {
    if(sys_prop.unx==1) s="find ";
    else                s="dir /a/b/s ";
  }

  so="";
  so=so+sys_prop.tmp+sys_prop.s+"out_ls_"+time_now_20;
  s=s+'"'+d+'"'+"  > "+'"'+so+'"';

  system(s.c_str());
  t.open(so.c_str());
  remove(so.c_str());

  int i, l, lf;
  char *a;

  // ..add the directory name to get a full name in some cases
  if(strlen(rec)==0)
  {
    ofstream out2(so.c_str());
    for(i=0; i<t.nl; i++)
    {
      lf=strlen(t.cl[i]);
      if(t.cl[i][lf-1]=='/')continue;
      if(t.cl[i][lf-1]=='.')continue;
      l=strlen(d);
      out2 << d;
      if(d[l-1] != sys_prop.s) out2 << sys_prop.s;
      out2 << t.cl[i] << endl;
    }
    out2.close(); 
    t.close();
    t.open(so.c_str());
    remove(so.c_str());
  }

  // .. set n and the pointers fn[]
  n=t.nl;
  fn=new char*[n+1];


  for(i=0; i<n; i++)
  {
    a=t.cl[i];
    if(d[0]=='.')fn[i]=&a[2];
    else         fn[i]=a;
  }
}

void mkwsystem(const char s[])
{
  if (quiet!=1)cout << "executing: " << s << flush;
  system(s);
  if (quiet!=1)cout << "  done" << endl;
}

int mkw_utime(const char fname[], time_t t)
{
  // .. modify the access and modification times of file fname[] to t

  time_t atime, mtime;
  atime=t;
  mtime=t;

  // .. create the structure used by utime()
  struct utimbuf a;
  a.actime=atime;
  a.modtime=mtime;

  // .. use utime()
  int i=utime(fname, &a);
  return i;
}

void mstrcpy(char a[], const char b[])
{
  // .. replacement for strcpy() to avoid microsoft compiler warnings
  for (int i=0; i<INT_MAX; i++)
  {
    if (&a[i]==NULL)return;
    if (&b[i]==NULL)return;
    a[i]=b[i];
    if (b[i]==0)return;
  }
}

void mstrncpy(char a[], const char b[], const int n)
{
  // .. replacement for strncpy() to avoid microsoft compiler warnings
  for (int i=0; i<n; i++)
  {
    if (&a[i]==NULL)return;
    if (&b[i]==NULL)return;
    a[i]=b[i];
    if (b[i]==0)return;
  }
}

int order_ascending_int(const void *aa, const void *bb)
{
  // .. used with qsort() to put an array of type int in ascending order
  const int& a=*(const int*)aa;
  const int& b=*(const int*)bb;
  return a-b;
}

void p2a(const char a[], char*& c, int& n)
{
  // .. a[] is a command with the output sent to the pipe
  // .. what is created is c[] of length n with the output

  FILE *p=POPEN(a, "r");

  if (p==NULL)
  {
    n=-1;
    return;
  }

  // .. get the size of the data in bytes as n
  const int NN=65536;
  char b[NN];
  int nr;
  n=0;
  do
  {
    nr=fread(b, 1, NN, p);
    n=n+nr;
    if (nr<NN)break;
  } while (n<INT_MAX);
  PCLOSE(p);

  // .. allocate enough space and run/read again
  // .. there is possibly a problem here if the 2nd run has more
  // .. output than the first run and hence allocate more space than
  // .. may usually be needed
  int m;
  m=NN;
  if (n>NN)m=2*n;
  c=new char[m+1];

  int i;
  if (n<NN)
    for (i=0; i<n; i++)c[i]=b[i];
  else
  {
    // .. re-run needed
    p=POPEN(a, "r");
    nr=fread(c, 1, m, p);
    PCLOSE(p);
  }
}

void replace(char s[], char a, char b)
{
  // .. change a by b in the s
  int i, l;
  l=strlen(s);
  for (i=0; i<l; i++)if (s[i]==a)s[i]=b;
}

void replace(string& s, char a, char b)
{
  // .. change a by b in the s
  int i, l;
  l=s.length();
  for (i=0; i<l; i++)if (s[i]==a)s[i]=b;
}

void replace_string(char c[], char d[], int& m, const char a[], const char b[])
{
  // .. create d[] of length m from c[]
  // .. by replacing c-string a[] by b[]
  // .. on input m indicates the capacity of d

  int la, lb, lc;
  la=strlen(a);
  lb=strlen(b);
  lc=strlen(c);

  // .. get the number of occurrences as num
  int i, j, k, r, nexti, num=0;
  for(i=0; i<=lc-la; i++)
  {
    k=0;
    for(j=i; j<i+la; j++)
      if(c[j]==a[j-i])k++;
      else           {k=0; break;}
    if(k<la)continue;
    num++;
  }

  // .. determine the size of d[] as mm and check that d[] has enough capacity
  int mm=la+num*(lb-la);
  if(mm>m)
  {
    cout << "line number=" << __LINE__ << " of file=" << __FILE__ << endl;
    cout << "m=" << m << " is less than mm=" << mm << endl;
    cout << "c[]=" << c << " a[]=" << a << "  b[]=" << b << endl;
    cout << "..not enough space to replace a[] by b[]" << endl;
    exit(0);
  }

  // .. create d[] doing the substitutions
  r=0;
  nexti=0;
  for(i=0; i<=lc-la; i++)
  {
    k=0;
    for(j=i; j<i+la; j++)
      if(c[j]==a[j-i])k++;
      else           {k=0; break;}
    // .. if k==la a match has been found
    if(k<la)
    {
      d[r]=c[i];
      r++;
      nexti=i+1;
      continue;
    }
    else
    {
      for(j=0; j<lb; j++)
      {
        d[r]=b[j];
        r++;
      }
      i=i+(la-1);
      nexti=i+1;
    }
  }
  for(i=nexti; i<lc; i++)
  {
    d[r]=c[i];
    r++;
  }
  d[r]=0;
  m=r+1;  
}

void replace_string(char c[], char d[], const char a[], const char b[])
{
  // .. create d[] from c[] by replacing c-string a[] by b[]
  // .. d[] must have enough capacity for the result

  char dd[1024];
  int i, m=1024;
  replace_string(c, dd, m, a, b);
  for(i=0; i<=m; i++)d[i]=dd[i];
}

string int_2_string(int m, int w)
{
  // return a string with the number m in a width of w
  int ww=w;
  if (w<=0)ww=3;
  if (ww>80)ww=80;
  string s;
  stringstream sstem;
  sstem.clear();
  sstem.str("");
  sstem << setw(ww) << setfill('0') << m;
  s=sstem.str();
  sstem.clear();
  return s;
}


void textfile(const char s[], char*& c, int& n, int& nl, char**& cl)
{
  // .. read a file to be treated as a textfile to create nl entries of cl[]
  int i;
  // .. read in the file
  c=f2a(s, n);

  // .. count the lines by counting the number of 10s and 13s
  nl=0;
  int nl10=0, nl13=0, ieol=10;
  for (i=0; i<n; i++)
  {
    if (c[i]==10)nl10++;
    if (c[i]==13)nl13++;
  }

  // .. deal with mis-match of number 13 chars by converting to char 10
  if(nl13>0)
  {
    for(i=0; i<n-1; i++)
    {
      if(c[i]==13 && c[i+1]!=10)
      {
        c[i]=10;
        nl10++;
      }
    }
  }

  // .. usually use 10 but use 13 if possibly a mac file
  nl=nl10;
  ieol=10;
  if (1<=nl13 && 4*nl10<=nl13)
  {
    nl=nl13;
    ieol=13;
  }

  // .. set the lines and make them null terminated
  cl=new char*[nl+1];

  // .. but do not do anything if n<=0 and take case of c[]
  if (n<0)c=new char[1];
  if (n<=0) return;

  // .. first line
  cl[0]=c;
  nl=1;

  // .. subsequent lines
  for (i=0; i<n-1; i++)
    if (c[i]==ieol)
    {
      cl[nl]=&c[i+1];
      nl++;
    }

  // .. add a char if the last line is not completed
  if (c[n-1]!=ieol)
  {
    c[n]=0;
    n++;
  }

  // .. ensure that all lines are null-terminated
  for (i=0; i<n; i++)
  {
    if (c[i]==13)c[i]=0;
    if (c[i]==10)c[i]=0;
  }
}

void textfile(const char s[], char*& c, int& nl, char**& cl)
{
  int n;
  textfile(s, c, n, nl, cl);
}

void textfile2(const char s[], char*& c, int& n, int& nl, char**& cl)
{
  textfile(s, c, n, nl, cl);
  cout << "File " << s << " has " << nl << " lines and " << n << " bytes"
       << endl;
}

void time_t_to_struct_tm(time_t& g, struct tm& t)
{
#ifdef _MSC_VER
#if(_MSC_VER>=1400)
  gmtime_s(&t, &g);
#else
  t=*gmtime(&g);
#endif
#else
  t=*gmtime(&g);
#endif
}


void unique_sort(const void *a, int n, int sz, 
                 int(*f)(const void*, const void*), int o[], int& m, int g[])
{
  // ..input:  a[i], i=0,1,...,n-1 is data which will not be changed
  // ..        each item requires sz bytes
  // ..        f() takes pointer-to-pointers as arguments
  // ..        and returns <0, 0 or >0 
  // ..        if arg1 should be before/the same/after arg2
  // ..output: o[], such that a[o[i]], i=0,1,... n-1 are in order
  // ..        according to the criteria of f()
  // ..        m<=n is the number of unique entries
  // ..        a[o[j]], j=g[i], ..., g[i+1]-1 are the same according to f()
  // ..        m+1 values of g[] are set with g[0]=0, g[m]=n

  // .. the first 4 args are the same as qsort()

  m=n;
  if(n<=0)return;

  int i, j;
  // .. re-cast to a pointer which can be used
  const char *b=(const char*)a;

  // ..create an array of pointers holding the addresses 
  const char **pb=new const char*[n+1];
  const char *pb0;
 
  for(i=0; i<n; i++)pb[i]=b+i*sz;

  // ..sort the addresses and get the order array o[]
  pb0=pb[0];
  qsort(pb, n, sizeof(pb[0]), f);
  for(i=0; i<n; i++)o[i]=(pb[i]-pb0)/sz;

  // ..now determine the number of different entries and create g[]
  // ..indicating where each different entry starts
  g[0]=0;
  m=0;
  for(i=1; i<n; i++)
    if( f((const void*)&pb[i-1], (const void*)&pb[i])!=0 )
    {
      m++;
      g[m]=i;
    }
  m++;
  g[m]=n;

  // ..sort each group of the same entry so as to preserve order
  for(i=0; i<m; i++)
  {
    j=g[i+1]-g[i];
    if(j==1)continue;
    qsort(&o[g[i]], j, sizeof(o[0]), order_ascending_int);
  }

  // .. release the space created
  delete[] pb;
}

streambuf *cout_backup=NULL;
streambuf *null_buffer=NULL;
ofstream  other_buffer;

void turn_off_cout()
{
  cout_backup=cout.rdbuf();
  other_buffer.open(sys_prop.nul);
  null_buffer=other_buffer.rdbuf();
  if(null_buffer!=NULL) cout.rdbuf(null_buffer);
}

void turn_on_cout()
{
  if(cout_backup!=NULL) cout.rdbuf(cout_backup);
}


